from ..layers import Layer
from .meta import achain

__all__ = [
  'skip_connection'
]

def skip_connection(*body):
  def op(*incoming):
    result = achain(*body)(*incoming)
    if isinstance(result, Layer):
      return tuple(incoming) + (result, )
    else:
      return tuple(incoming) + tuple(result)

  return op


from ..meta import derive
from .meta import Parameter
from .common import const_init

__all__ = [
  'scaled',
  'biased',
  'default_dependency_strategy'
]

def default_dependency_strategy(param):
  if param.properties.get('weights', False):
    return scaled(param, dependency_strategy=default_dependency_strategy)
  elif param.properties.get('biases', False):
    return biased(param, dependency_strategy=default_dependency_strategy)
  else:
    return param

class AugmentedParameter(Parameter):
  def __init__(self, param : Parameter, op, augment):
    self.op = op

    self.augments = [
      augment(var.shape)
      for var in param.variables()
    ]

    self.param = param
    super(AugmentedParameter, self).__init__(param.shape, properties=param.properties, name=param.name)

  def get_augments(self):
    return self.augments

  def variables(self):
    return self.param.variables()

  def get_output_for(self, *vars):
    return self.param.get_output_for(*(
        self.op(var, augment())
        for augment, var in zip(self.augments, vars)
    ))

  def apply(self):
    for var, augment in zip(self.param.variables(), self.augments):
      var.assign(self.op(var, augment()))

def _scale(var, augment):
  return var * augment

def _bias(var, augment):
  return var + augment

ScaledParameter = derive('ScaledParameter').based_on(AugmentedParameter).with_fixed(
  op=_scale,
  augment=lambda shape: const_init(value=1, augment=True)(shape=())
)

scaled = ScaledParameter

BiasedParameter = derive('BiasedParameter').based_on(AugmentedParameter).with_fixed(
  op=_bias,
  augment=lambda shape: const_init(value=0, augment=True)(shape=())
)

biased = BiasedParameter





from builtins import getattr

from ..meta import *

__all__ = [
  'Layer',
  'InputLayer',

  'FunctionalLayer',

  'custom_layer',

  'CompositeLayer',
  'composite',

  'propagate',
  'reduce_graph',
  'find_in_graph',

  'get_output',
  'get_all_outputs',

  'get_output_shape',
  'get_all_output_shapes',

  'get_parameters',
  'get_all_parameters',

  'get_variables',
  'get_all_variables',

  'get_layers',
  'get_inputs_layers',

  'model_selector',

  'model_from_class',
  'model_from',

  'modify_graph'
]


class Layer(object):
  def __init__(self, name=None):
    self.name = name

  def __str__(self):
    if not hasattr(self, 'name') or self.name is None:
      return self.__class__.__name__
    else:
      return self.name

  def __repr__(self):
    return str(self)

  def get_output_shape_for(self, *input_shapes, **kwargs):
    raise NotImplementedError

  def get_output_for(self, *args, **kwargs):
    raise NotImplementedError


class InputLayer(Layer):
  def __init__(self, shape, name=None):
    self.shape = shape

    super(InputLayer, self).__init__(name)

  def get_output_shape_for(self):
    return self.shape

  def get_output_for(self, *args, **kwargs):
    raise ValueError('Input was asked to return a value, this should not have happened.')


class FunctionalLayer(Layer):
  def __init__(self, *incoming, name=None):
    self.incoming = incoming
    super(FunctionalLayer, self).__init__(name)


class CustomLayer(FunctionalLayer):
  def __init__(self, f, shape_f, *incoming, name=None):
    super(CustomLayer, self).__init__(*incoming, name=name)
    self.f = f
    self.shape_f = shape_f

  def get_output_for(self, *args):
    return self.f(*args)

  def get_output_shape_for(self, *input_shapes):
    return self.shape_f(*input_shapes)

_default_shape_f = lambda *input_shapes: input_shapes[0]
custom_layer = lambda f, shape_f=_default_shape_f, name=None: lambda *incoming: CustomLayer(f, shape_f, *incoming, name=name)


class CompositeLayer(FunctionalLayer):
  def  __init__(self, *incoming, subnetwork, name=None):
    super(CompositeLayer, self).__init__(*incoming, name=name)

    self.subnetwork = subnetwork(*incoming)
    self.subnetwork_layers = get_subgraph(self.subnetwork, incoming)

    self.parameters = [
      param
      for layer in self.subnetwork_layers
      for param in getattr(layer, 'params', [])
    ]

  def get_output_for(self, *args, **kwargs):
    substitutes = dict(zip(self.incoming, args))
    return get_output(self.subnetwork, substitutes=substitutes, **kwargs)

  def get_output_shape_for(self, *input_shapes, **kwargs):
    substitutes = dict(zip(self.incoming, input_shapes))
    return get_output_shape(self.subnetwork, substitutes=substitutes, **kwargs)

composite = lambda subnetwork, name=None: lambda *incoming: CompositeLayer(*incoming, subnetwork=subnetwork, name=None)

def propagate(f, layers, substitutes=None):
  if substitutes is not None:
    known_results = dict(substitutes.items())
  else:
    known_results = dict()

  stack = list()
  stack.extend(layers)

  while len(stack) > 0:
    current_layer = stack.pop()

    if current_layer in known_results:
      continue

    incoming = getattr(current_layer, 'incoming', list())

    unknown_dependencies = [
      layer
      for layer in incoming
      if layer not in known_results
    ]

    if len(unknown_dependencies) == 0:
      known_results[current_layer] = f(
        current_layer,
        [ known_results[layer] for layer in incoming ]
      )

    else:
      stack.append(current_layer)
      stack.extend(unknown_dependencies)

  return known_results

def map_graph(f, layer_or_layers):
  """
  A wrapper over `propagate` for functions that does not depend on incoming values.
    Results are topologically ordered.

  :param f: `Layer -> value`
  :param layer_or_layers: layers (or a single layer) on which to compute `f` (including all dependencies).
  :return: topologically ordered outputs.
  """

  if isinstance(layer_or_layers, Layer):
    layers = [layer_or_layers]
  else:
    layers = layer_or_layers

  return list(propagate(
    f=lambda layer, *args: f(layer),
    layers=layers,
    substitutes=dict()
  ).values())

def reduce_graph(operator, strict=False):
  """
    Wraps operator into `propagate`-operator.

  :param operator: `Layer` -> function
  :param strict: if `False` use `apply_with_kwargs` wrapper on `operator` which filters key word arguments before passing
    them into propagated function; otherwise, passes `**kwargs` directly to the propagated function.
  :return: a getter, function `(list of layer, substitution dictionary=None, **kwargs) -> value`
    that computes the operator output for `layers`.
  """
  def getter(layers_or_layer, substitutes=None, **kwargs):
    from ..meta import apply_with_kwargs

    if isinstance(layers_or_layer, Layer):
      layers = [layers_or_layer]
    else:
      layers = layers_or_layer

    if substitutes is None:
      substitutes = dict()

    if strict:
      wrapped_operator = lambda layer, incoming: operator(layer)(*incoming, **kwargs)
    else:
      wrapped_operator = lambda layer, incoming: apply_with_kwargs(operator(layer), *incoming, **kwargs)

    results = propagate(wrapped_operator, layers, substitutes)

    if isinstance(layers_or_layer, Layer):
      return results[layers_or_layer]
    else:
      return [results[layer] for layer in layers]

  return getter

def find_in_graph(predicate, layers):
  visited = set()
  stack = list(layers)

  while len(stack) > 0:
    current = stack.pop()
    if current in visited:
      continue

    if predicate(current):
      return current

    incoming = getattr(current, 'incoming', list())
    stack.extend(incoming)

  return None

def modify_graph(f, layers):
  """
  This function dynamically changes graph with outputs `layers` by propagating operator `f`
  from outputs to the inputs. Operator `f` receives a layer and must output tuple `(layer replacement, subgraph)`,
  where:
    - layer_replacement: must be either:
      - None: to remove current layer from the graph;
      - a `craynn.layers.meta.FunctionalLayer` instance: to replace current layer with, the instance will be copied and have incoming overridden;
      - a layer model: a replacement for the current layer.
    - subgraph: list of layers for further propagation. Must be a subset of incoming layers of the original layer.

  This function allows, for example, to prune graph.

  *Note*: this function might make shallow copies of the layers and override `incoming` value.
    Inappropriate operator may result in invalid graph. For example, removing second convolution from
    `conv(8) -> conv(16) -> conv(32)` will leave the last convolution mismatched kernel.

  *Note*: inputs layers are never copied.
  """

  import copy

  def propagate_operator(layer):
    replacement, subgraph = f(layer)

    subgraph_replacement = [
      layer
      for incoming in subgraph
      for layer in propagate_operator(incoming)
    ]

    if replacement is None:
      return [
        layer
        for sg in subgraph_replacement
        for layer in propagate_operator(sg)
      ]
    elif callable(replacement):
      replacement = replacement(*subgraph_replacement)
      return [ replacement ]
    elif isinstance(replacement, Layer):
      if hasattr(replacement, 'incoming'):
        replacement = copy.copy(replacement)
        replacement.incoming = tuple(subgraph_replacement)
      else:
        replacement = replacement

      return [ replacement ]
    else:
      raise ValueError('operator return value is not understood: %s.' % replacement)


  return [ output for layer in layers for output in propagate_operator(layer) ]

def layer_output(layer, *inputs, **modes):
  param_values = [
    param() for param in getattr(layer, 'parameters', [])
  ]

  return apply_with_kwargs(layer.get_output_for, *param_values, *inputs, **modes)


get_output = reduce_graph(lambda layer: lambda *inputs, **modes: layer_output(layer, *inputs, **modes), strict=True)

def get_all_outputs(layer_or_layers, substitutes, **modes):
  operator = lambda layer, inputs: layer_output(layer, *inputs, **modes)

  if isinstance(layer_or_layers, Layer):
    layers = [layer_or_layers]
  else:
    layers = layer_or_layers

  return propagate(operator, layers, substitutes)

get_output_shape = reduce_graph(lambda layer: layer.get_output_shape_for, strict=False)

get_all_output_shapes = lambda layers, substitutes=None, **kwargs: propagate(
  lambda layer, args: apply_with_kwargs(layer.get_output_shape_for, *args, **kwargs),
  layers, substitutes=substitutes
)

get_layers = lambda layers_or_layer: map_graph(lambda layer: layer, layers_or_layer)

get_inputs_layers = lambda layers_or_layer: [
  layer
  for layer in get_layers(layers_or_layer)
  if len(getattr(layer, 'incoming', [])) == 0
]

def get_subgraph(layer_or_layers, origins):
  if isinstance(layer_or_layers, Layer):
    layers = [layer_or_layers]
  else:
    layers = layer_or_layers


  return [
    layer

    for layer, is_origin in propagate(
      lambda layer, *args, **kwargs: False,
      layers, substitutes=dict([
        (origin, True) for origin in origins
      ])
    ).items()

    if not is_origin
  ]

def get_parameters(layer: Layer, **properties):
  """
  Get parameters that satisfy all `properties`.
  A parameter satisfies a property `prop = value` if:
    - value is None;
    - the parameter has property `prop` and its value equals to `value` or
    - the parameter lacks property `prop` and `value = False`.

  Note, that `prop = None` matches all parameters, this is done to
  place commonly used properties to default arguments and enable autocomplete for them.

  Parameters
  ----------
  layer layer parameters of which are considered.
  properties list of properties

  Returns
  -------
  list of parameters
  """
  parameters = getattr(layer, 'parameters', [])
  check_props = lambda param: all([
    (param.properties.get(k, False) == v) or (v is None)
    for k, v in properties.items()
  ])

  return [
    param
    for param in parameters
    if check_props(param)
  ]

def get_all_parameters(layer_or_layers, **properties):
  collected_params = map_graph(lambda layer: get_parameters(layer, **properties), layer_or_layers)
  return [ param for params in collected_params for param in params ]

def get_variables(layer: FunctionalLayer, **properties):
  """
  Collects all free variables from parameters with `properties` (see `get_params` for details).

  Returns
  -------
  list of free variables
  """

  return [
    variable
    for param in get_parameters(layer, **properties)
    for variable in param.variables()
  ]

def get_all_variables(layer_or_layers, **properties):
  collected_variables = map_graph(lambda layer: get_variables(layer, **properties), layer_or_layers)
  return [ param for params in collected_variables for param in params ]


def model_selector(criterion):
  """Decorator, changes signature and inserts checks into a layer model selector.

  This is a wrapper which inserts a common procedures for a selector:
  - signature checks for each model (must all be the same);
  - binding of model parameters;
  - replacement of selector signature by models' shared signature.

  Model selector is a meta layer model that selects a particular layer model from a provided list
  based on properties of incoming layer, i.e. defers selection of a model until network construction.

  Type of the selector: `list of models` -> `incoming layer` -> `layer`.

  Parameters
  ----------
  criterion : Selector
    selector to modify. This function can assume that models have the same signature (i.e. accept the same parameters).

  Returns
  -------
  Selector
    Selector with changed signature and validity checks.

  Examples
  -------
  Selecting model with proper dimensionality for convolutional layer
  based on dimensionality of the incoming layer:

  >>> @model_selector
  >>> def dimensionality_selector(models):
  >>>   def common_model(incoming):
  >>>     ndim = get_output_shape(incoming) - 2
  >>>     return models[ndim]
  >>>   return common_model
  """
  from inspect import signature

  def selector(models):
    assert len(models) > 0
    models_signatures = [
      signature(model) for model in models
      if model is not None
    ]

    if len(set(models_signatures)) != 1:
      pretty_signatures = '\n  '.join([ str(signature) for signature in set(models_signatures)])
      raise ValueError('All models must have the same signature, got:%s' % pretty_signatures)

    common_signature = models_signatures[0]

    bound_criterion = criterion(models)

    def common_model(*args, **kwargs):
      common_signature.bind(*args, **kwargs)

      def model(incoming):
        selected_model = bound_criterion(incoming)
        if selected_model is None:
          raise ValueError('Invalid incoming layer!')

        layer = selected_model(*args, **kwargs)(incoming)
        return layer

      return model

    common_model.__signature__ = common_signature
    return common_model

  return selector

def model_from_class(clazz, fixed, defaults):
  import inspect
  signature = inspect.signature(clazz)

  if 'incoming' in signature.parameters.keys():
    return carry(clazz, fixed, defaults, carried_arguments=['incoming'])
  elif 'incomings' in signature.parameters.keys():
    return carry(clazz, fixed, defaults, carried_arguments=['incomings'])

def model_from(clazz):
  import inspect
  signature = inspect.signature(clazz)

  if 'incoming' in signature.parameters.keys():
    return CarringExpression(clazz, carried=['incoming'])
  elif 'incomings' in signature.parameters.keys():
    return CarringExpression(clazz, carried=['incomings'])
  else:
    return CarringExpression(clazz, carried=[])



import tensorflow as tf

from ..reducers import momentum_average

__all__ = [
  'gradient_step',
  'momentum_step',
  'nesterov_step',
]

class GradientStep(object):
  def __init__(self, *var_spec):
    pass

  def __call__(self, grads):
    return grads

def gradient_step():
  def step(var_spec):
    return GradientStep(var_spec)

  return step

class MomentumStep(momentum_average):
  def __init__(self, var_spec, rho=0.99, initial_value=None):
    super(MomentumStep, self).__init__(*var_spec, rho=rho, initial_value=initial_value)

def momentum_step(rho=0.99):
  def step(var_spec):
    return MomentumStep(var_spec, rho=rho)

  return step

class NesterovStep(object):
  def __init__(self, var_spec, rho=0.99):
    self.momentum = momentum_average(*var_spec, rho)

  def __call__(self, grads):
    return [
      g + self.momentum.rho * m
      for g, m in zip(grads, self.momentum(grads))
    ]

def nesterov_step(rho=0.99):
  def step(var_spec):
    return NesterovStep(var_spec, rho)

  return step




from inspect import *

__all__ = [
  'get_signature',
  'save_parameters', 'set_parameters', 'load_parameters',
  'eval_network', 'NetworkMode'
]

class NetworkMode(object):
  def __init__(self, nn, modes):
    self.nn = nn
    self.modes = modes

  def __call__(self, *args, **kwargs):
    return self.nn.mode_call(args, kwargs, **self.modes)

  def mode(self, **modes):
    modes.update(self.modes)
    return NetworkMode(self.nn, modes)

def get_signature(inputs):
  return Signature(
    parameters=[
      Parameter(name=input.name, kind=Parameter.POSITIONAL_OR_KEYWORD)
      for input in inputs
    ]
  )

def set_parameters(session, nn, values : list):
  import tensorflow as tf
  vars = nn.variables()

  session.run(tf.group([
    tf.assign(var, value)
    for var, value in zip(vars, values)
  ]))

def load_parameters(session, nn, path : str):
  import pickle

  with open(path, 'rb') as f:
    values = pickle.load(f)

  return set_parameters(session, nn, values)

def save_parameters(session, nn, path : str):
  import pickle

  vars = nn.variables()
  values = session.run(vars)

  with open(path, 'wb') as f:
    pickle.dump(values, f)


def eval_network_fixed_batch(session, graph, outputs, substitutes, batch_size=1):
  from ..utils import traverse

  tensor_substitutes = dict([
    (graph[layer], value)
    for layer, value in substitutes.items()
  ])

  tensor_outputs = [
    graph[layer] for layer in outputs
  ]

  return traverse(session, tensor_outputs, tensor_substitutes.values(), tensor_substitutes.keys(), batch_size=batch_size)


def eval_network(session, graph, outputs, substitutes, batch_size=1):
  return eval_network_fixed_batch(session, graph, outputs, substitutes, batch_size=batch_size)



import tensorflow as tf

from ...meta import derive
from .meta import GradientOptimizer

__all__ = [
  'SGD', 'sgd',
  'Momentum', 'momentum',
  'RMSProp', 'rmsprop',
  'Adam', 'adam',
  'Adamax', 'adamax',
  'Nadam', 'nadam',
  'AMSGrad', 'amsgrad',
  'Nesterov', 'nesterov'
]

class TensorflowOptimizerWrapper(GradientOptimizer):
  def __init__(self, target, variables, tf_opt, opt_kwargs=None):
    if opt_kwargs is None:
      opt_kwargs = dict()

    self.tf_optimizer = tf_opt(**opt_kwargs)
    
    super(TensorflowOptimizerWrapper, self).__init__(target, variables)

  def apply_gradients(self, gradients):
    self.tf_optimizer.apply_gradients(zip(gradients, self.variables))


SGD = derive('SGD').based_on(TensorflowOptimizerWrapper).with_fixed(tf_opt=tf.keras.optimizers.SGD)

def sgd(learning_rate=1e-3):
  def optimizer(target, variables):
    return SGD(target, variables, opt_kwargs=dict(learning_rate=learning_rate))

  return optimizer


Momentum = derive('Momentum').based_on(TensorflowOptimizerWrapper).with_fixed(tf_opt=tf.keras.optimizers.SGD)

def momentum(learning_rate=1.0e-3, rho=0.9, eps=1e-7):
  def optimizer(target, variables):
    return Momentum(target, variables, opt_kwargs=dict(learning_rate=learning_rate, momentum=rho, eps=eps))

  return optimizer


RMSProp = derive('RMSProp').based_on(TensorflowOptimizerWrapper).with_fixed(tf_opt=tf.keras.optimizers.RMSprop)

def rmsprop(learning_rate=1.0e-3, rho=0.9, eps=1e-7):
  def optimizer(target, variables):
    return RMSProp(target, variables, opt_kwargs=dict(learning_rate=learning_rate, rho=rho, epsilon=eps))
  return optimizer


Adam = derive('Adam').based_on(TensorflowOptimizerWrapper).with_fixed(tf_opt=tf.keras.optimizers.Adam)

def adam(learning_rate=1.0e-3, beta1=0.9, beta2=0.999, eps=1e-7):
  def optimizer(target, variables):
    return Adam(target, variables, opt_kwargs=dict(
      learning_rate=learning_rate, beta_1=beta1, beta_2=beta2, epsilon=eps
    ))
  return optimizer


Adamax = derive('Adamax').based_on(TensorflowOptimizerWrapper).with_fixed(tf_opt=tf.keras.optimizers.Adamax)

def adamax(learning_rate=1.0e-3, beta1=0.9, beta2=0.999, eps=1e-7):
  def optimizer(target, variables):
    return Adamax(target, variables, opt_kwargs=dict(
      learning_rate=learning_rate, beta_1=beta1, beta_2=beta2, epsilon=eps
    ))
  return optimizer


Nadam = derive('Nadam').based_on(TensorflowOptimizerWrapper).with_fixed(tf_opt=tf.keras.optimizers.Nadam)

def nadam(learning_rate=1.0e-3, beta1=0.9, beta2=0.999, eps=1e-7):
  def optimizer(target, variables):
    return Nadam(target, variables, opt_kwargs=dict(
      learning_rate=learning_rate, beta_1=beta1, beta_2=beta2, epsilon=eps
    ))
  return optimizer


AMSGrad = derive('AMSGrad').based_on(TensorflowOptimizerWrapper).with_fixed(tf_opt=tf.keras.optimizers.Adam)

def amsgrad(learning_rate=1.0e-3, beta1=0.9, beta2=0.999, eps=1e-7):
  def optimizer(target, variables):
    return Adam(target, variables, opt_kwargs=dict(
      learning_rate=learning_rate, beta_1=beta1, beta_2=beta2, amsgrad=True, epsilon=eps
    ))
  return optimizer


Nesterov = derive('Nesterov').based_on(TensorflowOptimizerWrapper).with_fixed(tf_opt=tf.keras.optimizers.SGD)

def nesterov(learning_rate=1.0e-3, rho=0.9, eps=1e-7):
  def optimizer(target, variables):
    return Nesterov(target, variables, opt_kwargs=dict(
      learning_rate=learning_rate, momentum=rho, nesterov=True, eps=eps
    ))
  return optimizer


import tensorflow as tf

from ..meta import get_output_shape, model_selector

__all__ = [
  'get_kernel_size',
  'get_data_format',
  'get_padding',
  'get_pool_mode',
  'get_upscale_mode',

  'get_conv_op',
  'get_deconv_op',
  'get_pool_op',

  'get_channel_dim',
  'get_spatial_dims',

  'get_conv_output_shape',
  'get_deconv_output_shape',

  'to_channel_last',
  'to_channel_fist',

  'select_by_dim'
]

DATA_FORMATS = {
  1 : 'NCW',
  2 : 'NCHW',
  3 : 'NCDHW'
}

CHANNEL_LAST_DATA_FORMATS = {
  1 : 'NWC',
  2 : 'NHWC',
  3 : 'NDHWC'
}


def to_channel_last(tensor, ndim=None):
  if ndim is None:
    ndim = len(tf.shape(tensor)) - 2

  permutation = (0, ) + tuple(range(2, ndim + 2)) + (1, )
  return tf.transpose(tensor, permutation)

def to_channel_fist(tensor, ndim=None):
  if ndim is None:
    ndim = len(tf.shape(tensor)) - 2

  permutation = (0, ndim + 1) + tuple(range(1, ndim + 1))
  return tf.transpose(tensor, permutation)

def get_data_format(ndim, channel_last=False):
  if ndim in DATA_FORMATS:
    if channel_last:
      return CHANNEL_LAST_DATA_FORMATS[ndim]
    else:
      return DATA_FORMATS[ndim]
  else:
    raise ValueError('%d-dim instances format is not supported' % ndim)

POOL_MODES = {
  'max' : 'MAX',
  'mean' : 'AVG'
}

def get_pool_mode(mode):
  if mode is POOL_MODES:
    return POOL_MODES[mode]
  else:
    raise ValueError('Unknown pool mode %s' % (mode, ))

PADDINGS = {
  'valid' : 'VALID',
  'same' : 'SAME'
}

def get_padding(pad):
  if pad in PADDINGS:
    return PADDINGS[pad]
  else:
    return pad

UPSCALE_MODES = {
  'nearest' : tf.image.ResizeMethod.NEAREST_NEIGHBOR,
  'bilinear' : tf.image.ResizeMethod.BILINEAR,
  'bicubic' : tf.image.ResizeMethod.BICUBIC,
}

def get_upscale_mode(mode):
  if mode in UPSCALE_MODES:
    return UPSCALE_MODES[mode]
  else:
    raise ValueError('Unknown upscale mode %s' % (mode, ))

def conv1d(input, filters, strides, padding, data_format='NCW', dilations=None, name=None):
  return tf.nn.conv1d(
    input=input,
    filters=filters,
    stride=strides,
    padding=padding,
    data_format=data_format,
    dilations=dilations,
    name=name
  )

CONV_OPS = {
  1 : conv1d,
  2 : tf.nn.conv2d,
  3 : tf.nn.conv3d,
}

def get_conv_op(ndim):
  if ndim in CONV_OPS:
    return CONV_OPS[ndim]
  else:
    raise ValueError('%d-dim conv is not supported' % ndim)

DECONV_OPS = {
  1 : tf.nn.conv1d_transpose,
  2 : tf.nn.conv2d_transpose,
  3 : tf.nn.conv3d_transpose
}

def get_deconv_op(ndim):
  if ndim in DECONV_OPS:
    return DECONV_OPS[ndim]
  else:
    raise ValueError('%d-dim deconv is not supported' % ndim)

POOL_OPS = {
  ('max', 1) : tf.nn.max_pool1d,
  ('max', 2) : tf.nn.max_pool2d,
  ('max', 3) : tf.nn.max_pool3d,

  ('mean', 1) : tf.nn.avg_pool1d,
  ('mean', 2) : tf.nn.avg_pool2d,
  ('mean', 3) : tf.nn.avg_pool3d,
}

def get_pool_op(mode, ndim):
  if (mode, ndim) in POOL_OPS:
    return POOL_OPS[(mode, ndim)]
  else:
    raise ValueError('Uknown pooling operation: %s %dd' % (mode, ndim))


def get_kernel_size(kernel_size, ndim):
  if type(kernel_size) is int:
    return (kernel_size, ) * ndim
  else:
    try:
      if all([ type(n) for n in kernel_size ]) and len(kernel_size) == ndim:
        return tuple(kernel_size)
    except:
      raise ValueError('kernel_size is neither iterable, nor int [%s]' % kernel_size)

get_channel_dim = lambda shape: shape[1]
get_spatial_dims = lambda shape: shape[2:]

def int_ceil(a, b):
  return (a + b - 1) // b

def get_conv_output_shape(input_shape, kernel_size, stride=1, pad='valid', dilation=1):
  if input_shape is None:
    return None

  effective_kernel_size = (kernel_size - 1) * dilation + 1

  pad = pad.lower()

  if pad == 'same':
    effective_input_shape = input_shape
  elif pad == 'valid':
    effective_input_shape = input_shape - effective_kernel_size + 1
  else:
    raise ValueError('Unknown pad [%s]' % pad)

  return int_ceil(effective_input_shape, stride)

def get_deconv_output_shape(input_shape, kernel_size, stride=1, pad='valid'):
  if input_shape is None:
    return None

  pad = pad.lower()

  if pad == 'same':
    delta = kernel_size // 2
  elif pad == 'valid':
    delta = 0
  else:
    raise ValueError('Unknown pad [%s]' % pad)

  return (input_shape - 1) * stride - 2 * delta + kernel_size

@model_selector
def select_by_dim(models):
  def selector(incoming):
    ndim = len(get_output_shape(incoming)) - 2
    return models[ndim]

  return selector


from .meta import Dataset

__all__ = [
  'MappedDataset', 'ZippedDataset'
]

class MappedDataset(Dataset):
  def __init__(self, dataset : Dataset, f, batch_size=1):
    super(MappedDataset, self).__init__()

    self.f = f
    self.dataset = dataset
    self.batch_size = batch_size

  def _get_single(self, item):
    result = self._get_slice(
      slice(item, item + 1) if item != -1 else slice(item, None)
    )

    return tuple(
      r.reshape(r.shape[1:])
      for r in result
    )

  def _get_slice(self, item):
    result = self.f(
      *self.dataset._get_slice(item)
    )

    return result if isinstance(result, tuple) else (result, )

  def _get_sparse(self, item):
    result = self.f(
      *self.dataset._get_sparse(item)
    )

    return result if isinstance(result, tuple) else (result,)

  def get_subset(self, item):
    return self.dataset.subset(item).map(self.f)

  def size(self):
    return self.dataset.size()

  def data(self):
    from .utils import xmap

    result = xmap(
      self.f,
      self.dataset.indexed_seq(batch_size=self.batch_size),
      size=len(self.dataset)
    )

    return result if isinstance(result, tuple) else (result,)

  def shapes(self):
    size = len(self.dataset)
    probe = self._get_slice(slice(0, 1))

    return tuple(
      (size, ) + p.shape[1]
      for p in probe
    )


class ZippedDataset(Dataset):
  def __init__(self, first : Dataset, second : Dataset):
    super(ZippedDataset, self).__init__()

    self.first = first
    self.second = second

  def _get_single(self, item):
    return self.first._get_single(item) + self.second._get_single(item)

  def _get_slice(self, item):
    return self.first._get_slice(item) + self.second._get_slice(item)

  def _get_sparse(self, item):
    return self.first._get_sparse(item) + self.second._get_sparse(item)

  def get_subset(self, item):
    return self.first.subset(item).zip(self.second.subset(item))

  def size(self):
    return self.first.size()

  def data(self):
    return self.first.data() + self.second.data()

  def shapes(self):
    return self.first.shapes() + self.second.shapes()


from .meta import achain
from ..layers import find_in_graph

__all__ = [
  'with_inputs',
  'select', 'seek', 'nothing'
]

### Surprisingly, `nothing` is quite useful
nothing = lambda *incoming: incoming

def _selector(*items, search_subgraph=False):
  def selector(*incomings):
    results = []

    for item in items:
      if isinstance(item, int):
        results.append(incomings[item])
      elif isinstance(item, slice):
        results.extend(incomings[item])
      elif isinstance(item, str):
        if search_subgraph:
          layer = find_in_graph(lambda layer: getattr(layer, 'name', None) == item, incomings)
          if layer is None:
            raise IndexError('There are not any layers named %s in the incoming subgraph.' % item)

          results.append(layer)
        else:
          matches = [incoming for incoming in incomings if incoming.name == item]
          if len(matches) == 0:
            raise IndexError('There are not any layers named %s among inputs' % item)
          else:
            results.extend(matches)
      else:
        raise ValueError('Items must be intergers, slices or layer names, got %s' % (item, ))


    return results
  return selector

def _with_inputs(*items, search_subgraph=False):
  def contructor(*definition):
    op = achain(*definition)

    def subnet(*incomings):
      selected_incomings = _selector(*items, search_subgraph=search_subgraph)(*incomings)
      is_selected = [
        incoming in selected_incomings
        for incoming in incomings
      ]

      op_results = op(*selected_incomings)

      if not isinstance(op_results, list) and not isinstance(op_results, tuple):
        op_results = (op_results, )
      else:
        op_results = tuple(op_results)

      if len(op_results) < len(selected_incomings):
        op_results = op_results + (None, ) * (len(items) - len(op_results))

      op_results_indx = 0
      results = []

      for i in range(len(incomings)):
        if is_selected[i]:
          results.append(op_results[op_results_indx])
          op_results_indx += 1
        else:
          results.append(incomings[i])

      n = len(selected_incomings)
      results.extend(op_results[n:])

      return [
        result
        for result in results
        if result is not None
      ]

    return subnet
  return contructor

class WithInputs(object):
  def __init__(self, search_subgraph=False):
    self.search_subgraph = search_subgraph

  def __call__(self, *items):
    return _with_inputs(*items, search_subgraph=self.search_subgraph)

  def __getitem__(self, items):
    if not isinstance(items, tuple):
      items = (items, )

    return _with_inputs(*items, search_subgraph=self.search_subgraph)


with_inputs = WithInputs(search_subgraph=False)


class Select(object):
  def __init__(self, search_subgraph=False):
    self.search_subgraph = search_subgraph

  def __call__(self, *items):
    def constructor(*definition):
      return achain(
        _selector(*items, search_subgraph=self.search_subgraph),
        *definition
      )

    return constructor

  def __getitem__(self, items):
    if not isinstance(items, tuple):
      items = (items, )

    return self(*items)

select = Select(search_subgraph=False)
seek = Select(search_subgraph=True)


import tensorflow as tf

from .meta import FunctionalLayer, InputLayer
from .meta import model_from, derive
from .meta import find_in_graph

from ..parameters import glorot_uniform_init

__all__ = [
  'const_input', 'variable_input',
  
  'FunctionLayer', 'custom', 'nonlinearity',
  'ConcatLayer', 'concat',
  'FlattenLayer', 'flatten',
  'ReshapeLayer', 'reshape',
  'TransposeLayer', 'transpose',

  'ElementwiseLayer', 'ElementwiseSumLayer', 'ElementwiseMeanLayer',
  'ElementwiseMaxLayer', 'ElementwiseMinLayer',

  'elementwise', 'elementwise_sum', 'elementwise_mean',
  'elementwise_max', 'elementwise_min',

  'SquareDifference', 'square_difference'
]

class ConstInput(InputLayer):
  def __init__(self, const, name=None):
    self.value = tf.constant(const)
    super(ConstInput, self).__init__(shape=const.shape, name=name)

  def get_output_for(self):
    return self.value

class VariableInput(InputLayer):
  def __init__(self, shape, V=glorot_uniform_init(), name=None):
    self.parameters = [V(shape, trainable=True)]

    super(VariableInput, self).__init__(shape=shape, name=name)

  def get_output_for(self, V):
    return V

const_input = lambda const, name=None: \
  lambda : ConstInput(const, name=name)

variable_input = lambda shape, V=glorot_uniform_init(), name=None: \
  lambda : VariableInput(shape, V, name=name)


class FunctionLayer(FunctionalLayer):
  def __init__(self, f, *incoming, name=None):
    if name is None:
      name = f.__name__

    super(FunctionLayer, self).__init__(*incoming, name=name)

    self.f = f

  def get_output_for(self, *incoming):
    return self.f(*incoming)

  def get_output_shape_for(self, *input_shapes):
    return input_shapes[0]

custom = lambda f, name=None: lambda *incoming: FunctionLayer(f, *incoming, name=name)
nonlinearity = custom


class ConcatLayer(FunctionalLayer):
  def __init__(self, *incoming, axis=-1, name=None):
    super(ConcatLayer, self).__init__(*incoming, name=name)

    assert len(incoming) > 0

    self.axis = axis

  def get_output_for(self, *inputs):
    return tf.concat(values=inputs, axis=self.axis, name=str(self) + '_concat')

  def get_output_shape_for(self, *input_shapes):
    first = input_shapes[0]
    axis = (self.axis + len(first)) % len(first)

    return tuple([
      first[i] if i != axis else sum([s[i] for s in input_shapes])
      for i in range(len(first))
    ])


concat = lambda axis=-1, name=None: lambda *incoming: ConcatLayer(*incoming, axis=axis, name=name)

class ExpandConcatLayer(FunctionalLayer):
  def __init__(self, *incoming, name=None):
    super(ExpandConcatLayer, self).__init__(*incoming, name=name)
    assert len(incoming) > 0

  def get_output_for(self, *inputs):
    max_len = max([ len(input.shape) for input in inputs ])
    expanded = [
      tf.expand_dims(input, axis=-1) if len(input.shape) < max_len else input
      for input in inputs
    ]

    return tf.concat(expanded, axis=-1)

  def get_output_shape_for(self, *input_shapes):
    original = input_shapes[0]

    return original[:-1] + (sum([shape[-1] for shape in input_shapes]), )


expand_concat = lambda name=None: lambda *incoming: ExpandConcatLayer(*incoming, name=name)

from functools import reduce as _reduce

class FlattenLayer(FunctionalLayer):
  def __init__(self, incoming, outdim=2, name=None):
    super(FlattenLayer, self).__init__(incoming, name=name)

    self.outdim = outdim

  def get_output_for(self, incoming):
    in_shape = tf.shape(incoming)

    out_shape = tf.stack([
      in_shape[i] for i in range(self.outdim - 1)
    ] + [
      tf.reduce_prod(in_shape[self.outdim - 1:])
    ])

    return tf.reshape(incoming, out_shape)

  def get_output_shape_for(self, input_shapes):
    return input_shapes[:self.outdim - 1] + (
      _reduce(
        lambda a, b: a * b if a is not None and b is not None else None,
        input_shapes[self.outdim - 1:],
        1
      ),
    )

flatten = lambda outdim=2, name=None: lambda incoming: FlattenLayer(incoming, outdim=outdim, name=name)

class ReshapeLayer(FunctionalLayer):
  def __init__(self, incoming, new_shape, name=None):
    super(ReshapeLayer, self).__init__(incoming, name=name)

    assert len([dim for dim in new_shape if dim < 0]) < 2, 'ambiguous new shape'

    self.new_shape = tuple([ (-1 if s is None else s) for s in new_shape ])

  def get_output_for(self, incoming):
    return tf.reshape(incoming, self.new_shape, name=str(self) + '_reshape')

  def get_output_shape_for(self, input_shape):
    import numpy as np

    if -1 in self.new_shape:
      if all(dim is not None for dim in input_shape):
        total = np.prod(input_shape, dtype='int64')
        known_dims = np.prod([ dim for dim in self.new_shape if dim is not None], dtype='int64')
        assert total % known_dims == 0, 'can not broadcast %s into %s' % (input_shape, self.new_shape)
        inferred = total // known_dims

        return tuple(dim if dim is not None else inferred for dim in self.new_shape)
      else:
        return tuple(dim if dim >= 0 else None for dim in self.new_shape)

    else:
      return self.new_shape

reshape = lambda new_shape, name=None: lambda incoming: ReshapeLayer(incoming, new_shape, name=name)

class TransposeLayer(FunctionalLayer):
  def __init__(self, incoming, perm, name=None):
    super(TransposeLayer, self).__init__(incoming, name=name)
    self.perm = perm

  def get_output_shape_for(self, input_shape):
    return tuple([ input_shape[i] for i in self.perm ])

  def get_output_for(self, input):
    return tf.transpose(input, perm=self.perm)

transpose = lambda perm, name=None: lambda incoming: TransposeLayer(incoming, perm=perm, name=name)

class ElementwiseLayer(FunctionalLayer):
  def __init__(self, *incoming, op, name=None):
    super(ElementwiseLayer, self).__init__(*incoming, name=name)
    self.op = op

  def get_output_for(self, *inputs):
    return self.op(*inputs)

  def get_output_shape_for(self, *input_shapes):
    for shape in input_shapes[1:]:
      if tuple(shape) != tuple(input_shapes[0]):
        raise ValueError('An elementwise operation requires all input shapes to be the same!')

    return input_shapes[0]

elementwise = model_from(ElementwiseLayer)()


ElementwiseSumLayer = derive('ElementwiseSumLayer').based_on(ElementwiseLayer).with_fixed(
  op=lambda *inputs: sum(inputs)
)
elementwise_sum = model_from(ElementwiseSumLayer)()

ElementwiseMeanLayer = derive('ElementwiseMeanLayer').based_on(ElementwiseLayer).with_fixed(
  op=lambda *inputs: sum(inputs) / len(inputs)
)
elementwise_mean = model_from(ElementwiseMeanLayer)()

from functools import reduce

ElementwiseMaxLayer = derive('ElementwiseMaxLayer').based_on(ElementwiseLayer).with_fixed(
  op=lambda *inputs: reduce(tf.maximum, inputs)
)
elementwise_max = model_from(ElementwiseMaxLayer)()

ElementwiseMinLayer = derive('ElementwiseMinLayer').based_on(ElementwiseLayer).with_fixed(
  op=lambda *inputs: reduce(tf.minimum, inputs)
)
elementwise_min = model_from(ElementwiseMinLayer)()


class SquareDifference(FunctionalLayer):
  def __init__(self, incoming1, incoming2, axes=None, name=None):
    from .meta import get_output_shape

    super(SquareDifference, self).__init__(incoming1, incoming2, name=name)
    if axes is None:
      shape = get_output_shape(incoming1)
      self.axes = list(range(len(shape)))[1:]
    else:
      self.axes = axes

  def get_output_for(self, input1, input2):
    return tf.reduce_mean(
      (input1 - input2) ** 2,
      axis=self.axes
    )

  def get_output_shape_for(self, shape1, shape2):
    return [ s for i, s in enumerate(shape1) if i not in self.axes]

square_difference = lambda axes=None, name=None: lambda incoming1, incoming2: SquareDifference(
  incoming1, incoming2, axes=axes, name=name
)


import tensorflow as tf

from .meta import *

from ..parameters import zeros_init

__all__ = [
  'SoftmaxGate',
  'softmax_gate'
]

class SoftmaxGate(FunctionalLayer):
  def __init__(self, *incoming, w=zeros_init(), name=None):
    super(SoftmaxGate, self).__init__(*incoming, name=name)

    self.parameters = [
      w(shape=(len(incoming),), weights=True, trainable=True, name='w')
    ]

  def get_output_for(self, w, *inputs):
    stacked = tf.stack(inputs, axis=1)
    dim_def = [None, Ellipsis] + [None] * (len(inputs[0].shape) - 1)
    coefs = tf.nn.softmax(w)

    return tf.reduce_sum(stacked * coefs[dim_def], axis=1)

  def get_output_shape_for(self, *input_shapes):
    return input_shapes[0]

softmax_gate = model_from(SoftmaxGate)()


import numpy as np
import tensorflow as tf

from .meta import FunctionalLayer, get_output_shape

from ..parameters import zeros_init, ones_init
from ..nonlinearities import softplus

from ..utils import normalize_axis

class GaussianMixtureLayer(FunctionalLayer):
  def __init__(
    self, incoming, num_components,
    mean=zeros_init(), log_sigma=zeros_init(), logit_priors=zeros_init(),
    name=None, axis=-1
  ):
    input_shape = get_output_shape(incoming)
    self.axis = normalize_axis(len(input_shape), axis)

    self.parameters = [
      mean(
        shape=tuple(input_shape[a] for a in self.axis) + (num_components, ),
        name='mean',
        biases=True, trainable=True
      ),
      log_sigma(
        shape=(num_components,),
        name='sigma',
        scales=True, trainable=True
      ),
      logit_priors(
        shape=(num_components, ),
        name='logit_priors',
        trainable=True
      )
    ]

    self.mean_broadcast = tuple(
      None if a in self.axis else slice(None, None, None)
      for a in self.axis
    ) + (slice(None, None, None), )

    self.input_broadcast = (slice(None, None, None), ) * len(input_shape) + (None, )

    self.sigma_broadcast = (None, ) * (len(input_shape) - len(self.axis)) + (slice(None, None, None), )

    self.log_2_pi = tf.constant(np.log(2 * np.pi), dtype=tf.float32)

    super(GaussianMixtureLayer, self).__init__(incoming, name=name)

  def get_output_for(self, mean, log_sigma, logit_priors, X):
    dim = len(self.axis)
    sigma = tf.exp(log_sigma)
    log_priors = tf.nn.log_softmax(logit_priors)

    diff = 0.5 * tf.reduce_sum(
      (X[self.input_broadcast] - mean[self.mean_broadcast]) ** 2,
      axis=self.axis
    ) / (sigma[self.sigma_broadcast] ** 2)

    log_probs = -dim / 2 * self.log_2_pi - dim * log_sigma - diff

    return tf.reduce_logsumexp(log_priors + log_probs, axis=-1)

  def get_output_shape_for(self, input_shape):
    return tuple(
      d
      for i, d in enumerate(input_shape)
      if i not in self.axis
    )


import tensorflow as tf

__all__ = [
  'elbo_norm'
]

def elbo_norm(X_original, X_reconstructed, code_mean, code_std, sigma_reconstructed=1.0, beta=None, exact=False):
  """
  Return Evidence Lower Bound for normally distributed (z | X), (X | z) and z:
    P(z | X) = N(`code_mean`, `code_std`);
    P(X | z) = N(`X_reconstructed`, `sigma_reconstructed`);
    P(z) = N(0, 1).

  :param X_original: original sample
  :param X_reconstructed: a sample from P(X | z)
  :param code_mean: mean of P(z | X)
  :param code_std: variance of P(z | X)
  :param sigma_reconstructed: variance for reconstructed sample, i.e. X | z ~ N(X_original, sigma_reconstructed)
    If a scalar, `Var(X | z) = sigma_reconstructed * I`, if tensor then `Var(X | z) = diag(sigma_reconstructed)`
  :param beta: coefficient for beta-VAE
  :param exact: if true returns exact value of ELBO, otherwise returns rearranged ELBO equal to the original
    up to a multiplicative constant, possibly increasing computational stability for low `sigma_reconstructed`.

  :return: (rearranged) ELBO.
  """
  reconstruction_loss = tf.reduce_mean((X_original - X_reconstructed) ** 2)

  code_penalty = tf.reduce_mean(
   tf.reduce_sum(code_std ** 2 + code_mean ** 2 - 2 * tf.log(code_std), axis=1)
  )

  if beta is None:
    ### code_penalty above is missing 1/2 coefficient.
    beta = tf.constant(0.5, dtype='float32')
  else:
    beta = tf.constant(beta / 2, dtype='float32')

  if exact:
    if is_tensor(sigma_reconstructed):
      normalization = 1 / 2 / sigma_reconstructed ** 2
    else:
      normalization = tf.constant(1 / 2 / sigma_reconstructed ** 2, dtype='float32')

    return normalization * reconstruction_loss + beta * code_penalty
  else:
    if is_tensor(sigma_reconstructed):
      normalization = 2 * sigma_reconstructed ** 2
    else:
      normalization = tf.constant(2 * sigma_reconstructed ** 2, dtype='float32')

    return reconstruction_loss + beta * normalization * code_penalty



__all__ = [
  'chain', 'achain', 'repeat',
  'subnetwork'
]

def flatten(l):
  if isinstance(l, list):
    return [ flatten(x) for x in l ]
  else:
    return l

def _chain(incoming, definition):
  net = incoming

  for layer in definition:
    if hasattr(layer, '__iter__'):
      net = _chain(net, layer)
    elif layer is None:
      pass
    else:
      net = layer(net)

  return net

chain = lambda *definition: lambda incoming: _chain(incoming, definition)

def _achain(incoming, definition):
  net = incoming

  if not isinstance(net, tuple) and not isinstance(net, list):
    net = [net]

  if not hasattr(definition, '__iter__'):
    try:
      net = definition(*net)
    except Exception as e:
      raise Exception('An error occurred while try to apply %s to %s' % (definition, net)) from e
  elif isinstance(definition, list):
    results = []
    for op in definition:
      try:
        result = _achain(net, op)
      except Exception as e:
        raise Exception('An error occurred while try to apply %s to %s' % (op, net)) from e

      if isinstance(result, tuple) or isinstance(result, list):
        results.extend(result)
      else:
        results.append(result)

    net = results

  elif isinstance(definition, tuple):
    for op in definition:
      try:
        net = _achain(net, op)
      except Exception as e:
        raise Exception('An error occurred while try to apply %s to %s' % (op, net)) from e
  else:
    raise RuntimeError('Unknown chain definition: %s' % (definition, ))

  return net

achain = lambda *definition: lambda *incoming: _achain(incoming, definition)

repeat = lambda n: lambda *definition: achain(
  (achain(*definition), ) * n
)

def subnetwork(inputs, outputs):
  """
  Applies existing graph substituting inputs by a new set of incoming layers.
  This procedure make shallow copies of layers, thus all parameters are shared.
  """
  from ..layers.meta import propagate
  from copy import copy

  def model(*incoming):
    assert len(incoming) == len(inputs), 'wrong number of incoming layers'

    substitutes = dict(zip(inputs, incoming))
    def mutator(layer, layer_incoming):
      new_layer = copy(layer)

      if hasattr(new_layer, 'incoming'):
        new_layer.incoming = layer_incoming

      return new_layer

    mutated = propagate(mutator, outputs, substitutes=substitutes)
    return [ mutated[output] for output in outputs ]

  return model



import tensorflow as tf

from ..reducers import moving_average
from .meta import GradientOptimizer

__all__ = [
  'adadelta'
]

class Adadelta(GradientOptimizer):
  def __init__(self, target, variables, learning_rate=1e-3, rho=0.99, eps=1e-7):
    super(Adadelta, self).__init__(target, variables)

    self.learning_rate = tf.constant(learning_rate, dtype=tf.float32)

    self.deltas_acc = moving_average(*variables, rho=rho, initial_value=tf.ones)
    self.second_momentum = moving_average(*variables, rho=rho)

    self.eps = tf.constant(eps, dtype=tf.float32)

  def apply_gradients(self, gradients):
    scaling = self.second_momentum([g ** 2 for g in gradients])

    deltas = [
      tf.sqrt(delta / (scale + self.eps)) * grad
      for delta, scale, grad in zip(self.deltas_acc, scaling, gradients)
    ]

    self.deltas_acc([
      delta ** 2
      for delta in deltas
    ])

    for var, delta in zip(self.variables, deltas):
      var.assign_sub(self.learning_rate * delta)

def adadelta(learning_rate=1e-1, rho=0.99, eps=1e-7):
  def optimizer(target, variables):
    return Adadelta(target, variables, learning_rate=learning_rate, rho=rho, eps=eps)

  return optimizer


import tensorflow as tf

from ...meta import derive
from ..meta import FunctionalLayer, model_from

from .conv_utils import *

__all__ = [
  'Upscale2DLayer',
  'BilinearUpscale2DLayer',
  'BicubicUpscale2DLayer',

  'upscale_2d',
  'bilinear_upscale_2d',
  'bicubic_upscale',

  'upscale',
  'bilinear_upscale',
  'bicubic_upscale_2d'
]

class UpscaleLayer(FunctionalLayer):
  def __init__(self, incoming, ndim, kernel_size=2, mode='nearest', name=None):
    super(UpscaleLayer, self).__init__(incoming, name=name)

    assert ndim == 2, 'only 2D upsampling is supported'

    self.ndim = ndim
    self.kernel_size = get_kernel_size(kernel_size, ndim)
    self.mode = mode.lower()

  def get_output_for(self, incoming):
    incoming_shape = tf.shape(incoming)
    new_size = tuple(incoming_shape[i + 2] * self.kernel_size[i] for i in range(self.ndim))

    transposed = to_channel_last(incoming, ndim=self.ndim)

    resized = tf.image.resize(
      transposed, size=new_size, method=get_upscale_mode(self.mode)
    )

    return to_channel_fist(resized, ndim=self.ndim)

  def get_output_shape_for(self, input_shape):
    return (input_shape[0], input_shape[1]) + tuple([
      int(size * factor)
      for size, factor in zip(get_spatial_dims(input_shape), self.kernel_size)
    ])

Upscale2DLayer = derive('Upscale2DLayer').based_on(UpscaleLayer).with_fixed(ndim=2, mode='nearest')
BilinearUpscale2DLayer = derive('Upscale2DLayer').based_on(UpscaleLayer).with_fixed(ndim=2, mode='bilinear')
BicubicUpscale2DLayer = derive('Upscale2DLayer').based_on(UpscaleLayer).with_fixed(ndim=2, mode='bicubic')

upscale_2d = model_from(Upscale2DLayer)()
bilinear_upscale_2d = model_from(BilinearUpscale2DLayer)()
bicubic_upscale_2d = model_from(BicubicUpscale2DLayer)()

upscale = upscale_2d
bilinear_upscale = bilinear_upscale_2d
bicubic_upscale = bicubic_upscale_2d


from ..layers.noise_ops import gaussian_rv
from .meta import achain

__all__ = [
  'vae'
]

def vae(encoder, decoder, latent_op=gaussian_rv(name='latent')):
  def subnet(*incoming):
    code_params = achain(encoder)(*incoming)

    if len(code_params) != 2:
      raise Exception('VAE encoder must produce 2 outputs: mean and std of posterior code distribution.')

    mean, std = code_params
    code = latent_op(mean, std)
    reconstructed, = achain(decoder)(code)
    return reconstructed, mean, std

  return subnet


import numpy as np
import tensorflow as tf

__all__ = [
  'sliced_seq',
  'xeval',
  'xmap'
]

def sliced_seq(size, batch_size=None):
  if batch_size is not None:
    n_batches = size // batch_size + (1 if size % batch_size != 0 else 0)

    for i in range(n_batches):
      yield slice(i * batch_size, min((i + 1) * batch_size, size))
  else:
    for s in range(size):
      yield s

def xeval(f, indexed_seq, size):
  """
  Evaluates `f` on `variables` with mini-batches.
  `f` must preserve the first dimension of the arguments.

  If `f` returns a single tensor (e.g. `return a_tensor`, but not `return (a_tensor, )`),
  `tfeval` returns a single array. Otherwise, it returns a tuple of arrays, with elements
  corresponding to each return value.

  `batch_size` sets size of mini-batches.
  `batch_size=None` is equivalent to `batch_size=1`
  but tensors in mini-batches lack the batch dimension, e.g.:

  `tfeval(f, a, b, batch_size=1)` makes the following calls:
  ```
  f(a[0:1], b[0:1]),
  f(a[1:2], b[1:2]),
  ...
  ```

  while with `batch_size=None`:

  ```
  f(a[0], b[0]),
  f(a[1], b[1]),
  ...
  ```
  """
  results = None
  singular = False

  make_buffer = lambda xs, bdim=1: tuple([
    np.ndarray(shape=(size,) + x.shape[bdim:], dtype=x.dtype.as_numpy_dtype)
    for x in xs
  ])

  for indx, batch in indexed_seq:
    batch_results = f(*batch)

    if not isinstance(batch_results, tuple):
      batch_results = (batch_results, )
      singular = True

    if results is None:
      results = make_buffer(batch_results, 0 if isinstance(indx, int) else 1)

    for br, r in zip(batch_results, results):
      r[indx] = br

  if len(results) == 1 and singular:
    return results[0]
  else:
    return results

def xmap(f, indexed_seq, size):
  """
  Similar to `xeval` but returns a single or multiple instances of `tf.Variable` instead.
  """
  make_buffer = lambda rs, bdim=1: tuple(
    tf.Variable(
      initial_value=tf.zeros((size,) + r.shape[bdim:], dtype=r.dtype),
      shape=(None,) + r.shape[bdim:],
      dtype=r.dtype,
      validate_shape=False,
    )
    for r in rs
  )

  results = None
  singular = False
  broadcasts = None

  index = None
  index_broadcasts = list()

  for indx, batch in indexed_seq:
    batch_results = f(*batch)

    if not isinstance(batch_results, tuple):
      batch_results = (batch_results, )
      singular = True

    if results is None:
      results = make_buffer(batch_results, 0 if isinstance(indx, int) else 1)

      broadcasts = tuple(
        (None,) + tuple(slice(None, None, None) for _ in br.shape)
        for br in batch_results
      )

      index_broadcasts = tuple(
        (slice(None, None, None), ) + tuple(None for _ in br.shape[1:])
        for br in batch_results
      )

      if not isinstance(indx, int):
        index = tf.range(size)

    if isinstance(indx, int):
      for br, r, broadcast in zip(batch_results, results, broadcasts):
        r.scatter_nd_update(
          indices=[[indx]],
          updates=br[broadcast]
        )
    else:
      for br, r, broadcast in zip(batch_results, results, index_broadcasts):
        r.scatter_nd_update(
          indices=index[indx][broadcast],
          updates=br
        )

  if len(results) == 1 and singular:
    return results[0]
  else:
    return results



import tensorflow as tf

from craynn.utils.tf_utils import make_vars

__all__ = [
  'moving_average',
  'moving_sum',
  'momentum_average',
  'max_average',
]

class accumulator(object):
  def __init__(self, *var_spec, rho=0.99, initial_value=None):
    self.rho = tf.constant(rho, dtype=tf.float32, shape=())
    self.crho = tf.constant(1 - rho, dtype=tf.float32, shape=())

    self._acc = make_vars(var_spec, initial_value=initial_value, trainable=False)

  @tf.function
  def __call__(self, values):
    if isinstance(values, list) or isinstance(values, tuple):
      return tuple(
        acc.assign(self.update(acc, value))
        for acc, value in zip(self._acc, values)
      )
    else:
      acc, = self._acc
      value = values

      return acc.assign(self.update(acc, value))

  def __iter__(self):
    if isinstance(self._acc, list):
      return iter(self._acc)
    else:
      return iter([self._acc])

  def update(self, acc, value):
    raise NotImplementedError

  def accumulator(self):
    return self._acc

class moving_average(accumulator):
  def update(self, acc, value):
    return self.rho * acc + self.crho * value

class momentum_average(accumulator):
  def update(self, acc, value):
    return self.rho * acc + value

class moving_sum(object):
  def __init__(self, *var_spec, initial_value=None):
    self._acc = make_vars(var_spec, initial_value=initial_value, trainable=False)

  @tf.function
  def __call__(self, values):
    if isinstance(self._acc, list):
      return [
        acc.assign_add(value)
        for acc, value in zip(self._acc, values)
      ]
    else:
      value = values
      self._acc.assign_add(value)

  def __iter__(self):
    if isinstance(self._acc, list):
      return iter(self._acc)
    else:
      return iter([self._acc])

class max_average(accumulator):
  def update(self, acc, value):
    return tf.maximum(self.rho * acc, value)


import numpy as np
import tensorflow as tf

from ..meta import as_stateful

__all__ = [
  'running_avg_loss'
]

@as_stateful
def running_avg_loss(
  initial_loss=0.0, half_life_slow=1024, half_life_fast=128, rho=0.99
):
  def monitor(loss):
    beta1 = np.exp(-np.log(2) / half_life_slow)
    beta2 = np.exp(-np.log(2) / half_life_fast)

    beta1 = tf.convert_to_tensor(beta1, dtype=loss.dtype)
    beta2 = tf.convert_to_tensor(beta2, dtype=loss.dtype)

    if rho is not None:
      _rho = tf.convert_to_tensor(rho, dtype=loss.dtype)
    else:
      _rho = None

    average_loss_slow = tf.Variable(
      initial_value=initial_loss,
      trainable=False,
      dtype=loss.dtype
    )

    average_loss_fast = tf.Variable(
      initial_value=0.0,
      trainable=False,
      dtype=loss.dtype
    )

    diff = average_loss_slow - average_loss_fast

    monitors_updates = [
      tf.assign(average_loss_slow, beta1 * average_loss_slow + (1 - beta1) * loss),
      tf.assign(average_loss_fast, beta2 * average_loss_fast + (1 - beta2) * loss),
    ]

    state = [average_loss_slow, average_loss_fast]

    if rho is not None:
      average_loss_diff = tf.Variable(
        initial_value=initial_loss,
        trainable=False,
        dtype=loss.dtype
      )

      monitors_updates.append(
        tf.assign(average_loss_diff, _rho * average_loss_diff + (1 - _rho) * diff)
      )

      observable = average_loss_diff

      state.append(average_loss_diff)
    else:
      observable = diff

    update_op = tf.group(monitors_updates)

    return state, update_op, dict(average_loss_change=observable)
  return monitor


import numpy as np
import tensorflow as tf

__all__ = [
  'normalize_variable_spec',
  'make_vars'
]

def _norm_spec(spec):
  if isinstance(spec, tuple):
    if len(spec) == 2 and (isinstance(spec[1], str) or isinstance(spec[1], tf.dtypes.DType)):
      return spec
    else:
      if not all([ isinstance(s, int) for s in spec ]):
        raise ValueError(
          'variable spec should be either an instance of tf.Variable, '
          'a tuple of integers (dtype defaults to float32), or shape-dtype pair, not %s' % (spec, )
        )

      return (spec, tf.float32)
  elif isinstance(spec, tf.Variable):
    return (
      tuple(spec.shape.as_list()),
      spec.dtype
    )

def normalize_variable_spec(var_spec):
  return [
    _norm_spec(spec) for spec in var_spec
  ]

def make_vars(var_spec, initial_value=None, trainable=False):
  var_spec = normalize_variable_spec(var_spec)

  if initial_value is None:
    initial_value_f = lambda shape, dtype: tf.zeros(shape=shape, dtype=dtype)
  elif not callable(initial_value):
    initial_value_f = lambda shape, dtype: tf.ones(shape=shape, dtype=dtype) * initial_value
  else:
    initial_value_f = initial_value

  return [
    tf.Variable(initial_value=initial_value_f(shape, dtype=dtype), dtype=dtype, trainable=trainable)
    for (shape, dtype) in var_spec
  ]


import numpy as np

__all__ = [
  'normalize',
  'box',
  'onehot',
  'categorical_encoding',
  'categorical_to_numerical',
  'binary_encoding',
  'split'
]

def normalize(data_ref, center=True, scale=True, axis=(0, ), scale_eps=1e-3, inline=True):
  """
  (X - mean) / (std + eps)

  Statistics are computed for the first argument.

  Example:
    `data_train, data_test = normalize(data_train)(data_train, data_test)`

  Returns: transform function
  """

  mean = np.mean(data_ref, axis=axis) if center else None
  std = np.std(data_ref, axis=axis) + scale_eps if scale else None

  def transform(*data):
    if len(data) == 0:
      return tuple()

    broadcast = tuple(
      None if i in axis else slice(None, None, None)
      for i, _ in enumerate(data[0].shape)
    )
    results = data if inline else tuple([ d.copy() for d in data ])

    for d, out in zip(data, results):
      if center:
        centered = np.subtract(d, mean[broadcast], out=out)
      else:
        centered = d

      if scale:
        np.divide(centered, std[broadcast], out=out)

    return results

  return transform

def box(data_ref, inline=True, eps=1e-3):
  max = np.max(data_ref, axis=0)
  min = np.min(data_ref, axis=0)
  delta = np.maximum(max - min, eps)

  def transform(*data):
    if len(data) == 0:
      return tuple()

    broadcast = (None, ) + tuple(slice(None, None, None) for _ in data[0].shape[1:])
    results = data if inline else tuple([ d.copy() for d in data ])

    for d, out in zip(data, results):
      np.subtract(d, min[broadcast], out=out)
      np.divide(out, delta[broadcast], out=out)

    return results

  return transform


def onehot(y, n_classes=None, dtype='float32'):
  if y.dtype.kind != 'i':
    y = y.astype('int64')

  if n_classes is None:
    n_classes = np.max(y) + 1

  y_ = np.zeros(shape=(y.shape[0], n_classes), dtype=dtype)
  y_[np.arange(y.shape[0]), y] = 1

  return y_

def ceil_log2(n):
  i = 0; x = 1

  while n > x:
    x *= 2; i += 1

  return i

def categorical_encoding(*data, fixed=None):
  if fixed is None:
    encoding = dict()
  else:
    encoding = fixed.copy()

  current = 0

  for d in data:
    for item in d:
      if item not in encoding:
        while current in encoding.values():
          current += 1

        encoding[item] = current

  return encoding

def categorical_to_numerical(data, encoding=None, fixed=None, dtype='int32'):
  if encoding is None:
    encoding = categorical_encoding(data, fixed=fixed)

  result = np.ndarray(shape=(len(data),), dtype=dtype)

  for i, item in enumerate(data):
    result[i] = encoding[item]

  return result

def binary_encoding(y, dtype='float32', size=None):
  y = np.array(y)

  if size is None:
    size = size if size is not None else (np.max(y) + 1)

  n_bits = ceil_log2(size)
  power_2 = 2 ** np.arange(n_bits)[::-1]

  return np.where(
    np.mod(y[:, None] // power_2, 2) != 0,
    np.ones(shape=tuple(), dtype=dtype),
    np.zeros(shape=tuple(), dtype=dtype),
  )

def split(*data, split_ratios=0.8, seed=None):
  if len(data) == 0:
    return tuple()

  try:
    iter(split_ratios)
  except TypeError:
    split_ratios = (split_ratios, 1 - split_ratios)

  assert all([r >= 0 for r in split_ratios])

  size = len(data[0])

  split_ratios = np.array(split_ratios)
  split_sizes = np.ceil((split_ratios * size) / np.sum(split_ratios)).astype('int64')
  split_bins = np.cumsum([0] + list(split_sizes))
  split_bins[-1] = size

  rng = np.random.RandomState(seed)
  r = rng.permutation(size)

  result = list()

  for i, _ in enumerate(split_ratios):
    from_indx = split_bins[i]
    to_indx = split_bins[i + 1]
    indx = r[from_indx:to_indx]

    for d in data:
      if isinstance(d, np.ndarray):
        result.append(d[indx])
      else:
        result.append([ d[i] for i in indx ])

  return tuple(result)


from inspect import signature, Signature, Parameter, _empty

__all__ = [
  'derive',
  'carry',
  'bind',

  'CarringExpression',
]

def get_var_arguments(signature):
  var_arguments = [
    name for name, param in signature.parameters.items()
    if param.kind == Parameter.VAR_POSITIONAL
  ]

  if len(var_arguments) == 0:
    return None
  else:
    return var_arguments[0]

def bind(signature, **kwargs):
  resolved_args = []
  resolved_kwargs = {}

  it = iter(signature.parameters.items())

  for name, param in it:
    if name not in kwargs:
      break

    if param.kind in [Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD]:
      resolved_args.append(kwargs[name])
    elif param.kind == Parameter.VAR_POSITIONAL:
      resolved_args.extend(kwargs[name])
      break

  for name, param in it:
    if name in kwargs:
      resolved_kwargs[name] = kwargs[name]

  return tuple(resolved_args), resolved_kwargs

class From(object):
  def __init__(self, what, base_class):
    self.what = what
    self.base_class = base_class

  def with_fixed(self, **kwargs):
    return fix_params_in_a_class(self.what, self.base_class, kwargs)

class Deriver(object):
  def __init__(self, what):
    self.what = what

  def based_on(self, base_class):
    return From(self.what, base_class)

derive = Deriver

def fix_params_in_a_class(name, base_class, fixing):
  original_signature = signature(base_class.__init__)

  new_signature = Signature(
    parameters=[
      param
      for name, param in original_signature.parameters.items()
      if name not in fixing
    ],
    return_annotation=original_signature.return_annotation
  )

  def new_init(self, *args, **kwargs):
    arguments = new_signature.bind(self, *args, **kwargs)
    arguments.apply_defaults()

    updated_kwargs = arguments.arguments.copy()
    updated_kwargs.update(fixing)

    args, kwargs = bind(original_signature, **updated_kwargs)

    base_class.__init__(*args, **kwargs)


  new_init.__signature__ = new_signature

  return type(name, (base_class, ), {
    '__init__' : new_init
  })

def carry(original, fixed, defaults, carried_arguments):
  original_signature = signature(original)

  for k in defaults:
    if k not in original_signature.parameters:
      raise ValueError('Attempting to provide default value for invalid argument %s.' % k)

  for k in fixed:
    if k not in original_signature.parameters:
      raise ValueError('Attempting to fix value of invalid argument %s!' % k)

  for k in carried_arguments:
    if k not in original_signature.parameters:
      raise ValueError('Attempting to carry invalid argument %s!' % k)

  mutual_keys = set(defaults.keys()) & set(fixed.keys())
  if len(mutual_keys) > 0:
    raise ValueError('The same arguments are fixed and provided default value for: %s!' % list(mutual_keys))

  mutual_keys = set(fixed.keys()) & set(carried_arguments)
  if len(mutual_keys) > 0:
    raise ValueError('The same arguments are fixed and carried: %s!' % list(mutual_keys))

  mutual_keys = set(defaults.keys()) & set(carried_arguments)
  if len(mutual_keys) > 0:
    raise ValueError('The same arguments are carried and provided default value for: %s!' % list(mutual_keys))

  ### replacing default values
  original_sig_with_defaults = Signature(
    parameters=[
      param.replace(default=defaults.get(name, param.default))
      for name, param in original_signature.parameters.items()
    ],
    return_annotation=original_signature.return_annotation
  )

  carried_arguments_set = set(carried_arguments)

  constructor_signature = Signature(
    parameters=[
      param.replace(kind=Parameter.POSITIONAL_OR_KEYWORD)
      for name, param in original_sig_with_defaults.parameters.items()
      if name not in fixed and name not in carried_arguments_set
    ],
    return_annotation=original_sig_with_defaults.return_annotation
  )

  model_signature = Signature(
    parameters=[
      param
      for name, param in original_sig_with_defaults.parameters.items()
      if name in carried_arguments_set
    ],
    return_annotation=original_sig_with_defaults.return_annotation
  )

  def constructor(*args, **kwargs):
    ### checking if all arguments are specified
    constructor_arguments = constructor_signature.bind(*args, **kwargs)
    constructor_arguments.apply_defaults()

    updated_arguments = constructor_arguments.arguments.copy()

    for k, v in fixed.items():
      updated_arguments[k] = v

    def model(*model_args, **model_kwargs):
      model_arguments = model_signature.bind(*model_args, **model_kwargs)
      args, kwargs = bind(original_signature, **model_arguments.arguments, **updated_arguments)

      result = original(*args, **kwargs)
      return result

    model.__signature__ = model_signature

    return model

  fixed_arguments_doc = ', '.join([
    '%s=%s' % (k, v) for k, v in fixed.items()
  ])

  constructor.__doc__ = "%s(%s)%s" % (original.__name__, fixed_arguments_doc, ('\n%s' % original.__doc__) if original.__doc__ is not None else '')
  constructor.__signature__ = constructor_signature

  return constructor

class CarringExpression(object):
  """
  Allows nice syntax for `carry` method:
  carry(<function>, ['param1', 'param2']).with_fixed(param1=value1).with_defaults(param3=value3)()
  """
  def __init__(self, original, fixed=None, defaults=None, carried=None):
    self.original = original

    self.fixed = dict() if fixed is None else fixed
    self.defaults = dict() if defaults is None else defaults
    self.carried = dict() if carried is None else carried

  def __call__(self):
    return carry(self.original, self.fixed, self.defaults, self.carried)

  def with_defaults(self, **kwargs):
    self.defaults.update(kwargs)
    return self

  def with_fixed(self, **kwargs):
    self.fixed.update(kwargs)
    return self


__all__ = [
  'split_groups'
]

def split_groups(data, min_size):
  regrouped = []

  for group in data:
    if len(group) < 2 * min_size:
      regrouped.append(group)
    else:
      n_groups = len(group) // min_size
      group_size = len(group) // n_groups

      for i in range(n_groups - 1):
        regrouped.append(group[i * group_size:(i + 1) * group_size])

      regrouped.append(group[(n_groups - 1) * group_size:])

  return regrouped


import tensorflow as tf

__all__ = [
  'mse', 'mae',
  'gmm_neg_log_likelihood',
]

def mse(target, predictions, weights=None):
  assert len(target.shape) == len(predictions.shape), 'target and predictions have different dimensionality'

  if weights is None:
    return tf.reduce_mean((target - predictions) ** 2)
  else:
    return tf.reduce_mean(weights * (target - predictions) ** 2)

def mae(target, predictions, weights=None):
  assert len(target.shape) == len(predictions.shape), 'target and predictions have different dimensionality'

  if weights is None:
    return tf.reduce_mean(tf.abs(target - predictions))
  else:
    return tf.reduce_mean(weights * tf.abs(target - predictions))

def gmm_neg_log_likelihood(target, means, log_sigmas, logit_priors=None):
  sigmas = tf.exp(log_sigmas)
  se = ((target[:, None] - means) / sigmas) ** 2

  if logit_priors is None:
    neg_log_likelihoods = -tf.reduce_logsumexp(
      -log_sigmas - 0.5 * se,
      axis=1
    )
  else:
    neg_log_likelihoods = -tf.reduce_logsumexp(
      -log_sigmas - 0.5 * se + tf.nn.log_softmax(logit_priors),
      axis=1
    )

  return tf.reduce_mean(neg_log_likelihoods)


import tensorflow as tf

from .step import *
from .scaling import *
from .meta import GradientOptimizer

__all__ = [
  'sgd', 'momentum', 'nesterov',
  'rmsprop', 'maxprop',
  'adagrad'
]

class GenericGradientOptimizer(GradientOptimizer):
  def __init__(self, target, variables, step, scaling, learning_rate=1e-3):
    super(GenericGradientOptimizer, self).__init__(target, variables)

    self._step_op = step(variables)
    self._scaling_op = scaling(variables)

    self.learning_rate = tf.constant(learning_rate, dtype=tf.float32)

  def apply_gradients(self, gradients):
    step = self._step_op(gradients)
    scaling = self._scaling_op(gradients)

    for var, delta, scale in zip(self.variables, step, scaling):
      if scale is None:
        var.assign_sub(self.learning_rate * delta)
      else:
        var.assign_sub(self.learning_rate * delta / scale)


class SGD(GenericGradientOptimizer):
  def __init__(self, target, variables, learning_rate=1e-3):
    super(SGD, self).__init__(
      target, variables,
      step=gradient_step(),
      scaling=no_scaling(),
      learning_rate=learning_rate
    )

def sgd(learning_rate=1e-3):
  def optimizer(target, variables):
    return SGD(target, variables, learning_rate=learning_rate)

  return optimizer


class Momentum(GenericGradientOptimizer):
  def __init__(self, target, variables, learning_rate=1e-3, rho=0.99):
    super(Momentum, self).__init__(
      target, variables,
      step=momentum_step(rho),
      scaling=no_scaling(),
      learning_rate=learning_rate
    )

def momentum(learning_rate=1e-3, rho=0.99):
  def optimizer(target, variables):
    return Momentum(target, variables, learning_rate=learning_rate, rho=rho)

  return optimizer


class Nesterov(GenericGradientOptimizer):
  def __init__(self, target, variables, learning_rate=1e-3, rho=0.99):
    super(Nesterov, self).__init__(
      target, variables,
      step=nesterov_step(rho),
      scaling=no_scaling(),
      learning_rate=learning_rate
    )

def nesterov(learning_rate=1e-3, rho=0.99):
  def optimizer(target, variables):
    return Nesterov(target, variables, learning_rate=learning_rate, rho=rho)

  return optimizer

class RMSProp(GenericGradientOptimizer):
  def __init__(self, target, variables, learning_rate=1e-3, rho=0.99, eps=1e-7):
    super(RMSProp, self).__init__(
      target, variables,
      step=gradient_step(),
      scaling=l2_scaling(rho, eps),
      learning_rate=learning_rate,
    )

def rmsprop(learning_rate=1e-3, rho=0.99, eps=1e-7):
  def optimizer(target, variables):
    return RMSProp(target, variables, learning_rate=learning_rate, rho=rho, eps=eps)

  return optimizer

class MaxProp(GenericGradientOptimizer):
  def __init__(self, target, variables, learning_rate=1e-3, rho=0.99):
    super(MaxProp, self).__init__(
      target, variables,
      step=gradient_step(),
      scaling=max_scaling(rho),
      learning_rate=learning_rate
    )

def maxprop(learning_rate=1e-3, rho=0.99):
  def optimizer(target, variables):
    return MaxProp(target, variables, learning_rate=learning_rate, rho=rho)

  return optimizer

class Adagrad(GenericGradientOptimizer):
  def __init__(self, target, variables, learning_rate=1e-2):
    super(Adagrad, self).__init__(
      target, variables,
      step=gradient_step(),
      scaling=total_path_scaling(),
      learning_rate=learning_rate
    )

def adagrad(learning_rate=1e-3):
  def optimizer(target, variables):
    return Adagrad(target, variables, learning_rate=learning_rate)

  return optimizer


import tensorflow as tf

from .meta import Parameter, parameter_model
from .common import default_weight_init

__all__ = [
  'NoisyParameter',
  'noisy'
]

class NoisyParameter(Parameter):
  def __init__(self, shape, eps=1e-3, w=default_weight_init, name=None, **properties):
    super(NoisyParameter, self).__init__(shape=shape, properties=properties, name=name)

    self.w = w(
      shape=shape,
      **properties,
      name=(name + '_w') if name is not None else None
    )
    self.eps = tf.constant(eps, dtype=tf.float32, shape=())

  def get_output_for(self, w):
    return w + tf.random.normal(shape=tf.shape(w), stddev=self.eps, dtype=w.dtype)

  def variables(self):
    return self.w.variables()

noisy = parameter_model(NoisyParameter)



import numpy as np

__all__ = [
  'downsample',
  'resize'
]

def downsample(*factors):
  def f(imgs):
    from skimage.transform import downscale_local_mean
    new_shape = (imgs.shape[1] // factors[0], imgs.shape[2] // factors[1])
    result = np.ndarray(shape=(imgs.shape[0], ) + new_shape + (imgs.shape[3], ), dtype=imgs.dtype)

    for i in range(imgs.shape[0]):
      result[i] = downscale_local_mean(imgs[i], factors=factors + (1, ))

    return result

  return f

def resize(new_size, order=3, dtype='float32', **kwargs):
  def f(imgs):
    from skimage.transform import resize

    result = np.ndarray(shape=(imgs.shape[0], ) + new_size + (imgs.shape[3], ), dtype=dtype)

    for i in range(imgs.shape[0]):
      result[i] = resize(imgs[i].astype(dtype), output_shape=new_size, order=order, **kwargs)

    return result

  return f


from ..layers import get_all_parameters

__all__ = [
  'apply_regularization'
]

def apply_regularization(layer, reg, **properties):
  params = get_all_parameters(layer, **properties)
  regularizers = []
  variables = []

  for param in params:
    result = reg(param)
    if isinstance(result, tuple):
      r, vs = result
      regularizers.append(r)
      variables.extend(vs)
    else:
      regularizers.append(result)

  return sum(regularizers), variables



import tensorflow as tf

from .common import glorot_normal_init
from .meta import Parameter, parameter_model

__all__ = [
  'DecompositionParameter',
  'decomposition'
]

class DecompositionParameter(Parameter):
  def __init__(self, shape, n, w1=glorot_normal_init(), w2=glorot_normal_init(), properties=None, name=None):
    super(DecompositionParameter, self).__init__(shape=shape, properties=properties, name=name)

    shape1 = (shape[0], n) + shape[2:]
    shape2 = (n, shape[1]) + shape[2:]

    self.w1 = w1(
      shape=shape1,
      **self.properties,
      name=(name + '_w1') if name is not None else None
    )

    self.w2 = w2(
      shape=shape2,
      **self.properties,
      name=(name + '_w2') if name is not None else None
    )

  def get_output_for(self, w1, w2):
    return tf.tensordot(w1, w2, axes=[(1,), (0,)])

  def variables(self):
    return self.w1.variables() + self.w2.variables()

decomposition = parameter_model(DecompositionParameter)


import tensorflow as tf

__all__ = [
  'Dataset'
]

class Dataset(object):
  def __init__(self):
    ### allows nice `dataset.subset[:100] syntax`
    self.subset = SubsetConstructor(self)

  ### for a signle item
  def _get_single(self, item):
    raise NotImplementedError()

  ### for slices
  def _get_slice(self, item):
    raise NotImplementedError()

  ### for any index arrays
  def _get_sparse(self, item):
    raise NotImplementedError()

  def get_subset(self, item):
    from .subsets import SlicedSubset, IndexedSubset
    if isinstance(item, slice):
      return SlicedSubset(self, item)
    else:
      return IndexedSubset(self, item)

  def size(self):
    raise NotImplementedError()

  def data(self):
    """
    Returns tensors held by the dataset.
    Warning: may result in allocation of a new dataset.
    """
    raise NotImplementedError()

  def numpy(self):
    return tuple(
      d.numpy()
      for d in self.data()
    )

  def materialize(self):
    from .variable import variable_dataset
    return variable_dataset(*self.data())

  def shapes(self):
    raise NotImplementedError()

  def __getitem__(self, item):
    if isinstance(item, int):
      return self._get_single(item)
    elif isinstance(item, slice):
      return self._get_slice(item)
    else:
      return self._get_sparse(item)

  def __len__(self):
    return int(self.size())

  def seq(self, batch_size=1):
    from .utils import sliced_seq

    if batch_size is None:
      for i in range(len(self)):
        yield self._get_single(i)
    else:
      for indx in sliced_seq(self.size(), batch_size=batch_size):
        yield self._get_slice(indx)

  def indexed_seq(self, batch_size=1):
    from .utils import sliced_seq

    if batch_size is None:
      for i in range(len(self)):
        yield i, self._get_single(i)
    else:
      for indx in sliced_seq(self.size(), batch_size=batch_size):
        yield indx, self._get_slice(indx)

  def batch(self, batch_size=1):
    if batch_size is None:
      indx = tf.random.uniform(
        shape=(),
        dtype=tf.int32,
        minval=0,
        maxval=self.size()
      )

      return self._get_single(indx)
    else:
      indx = tf.random.uniform(
        shape=(batch_size, ),
        dtype=tf.int32,
        minval=0,
        maxval=self.size()
      )

      return self._get_sparse(indx)

  def eval(self, f=None, batch_size=1):
    if f is None:
      return self.numpy()
    else:
      return self.map(f, batch_size=batch_size).numpy()

  def map(self, f, batch_size=1):
    from .common import MappedDataset
    return MappedDataset(self, f, batch_size=batch_size)

  def zip(self, other):
    from .common import ZippedDataset
    return ZippedDataset(self, other)

class SubsetConstructor(object):
  def __init__(self, dataset : Dataset):
    self.dataset = dataset

  def __getitem__(self, item):
    return self(item)

  def __call__(self, item):
    if isinstance(item, int):
      item = slice(item, item + 1,) if item != -1 else slice(item, None)

    return self.dataset.get_subset(item)


import tensorflow as tf

__all__ = [
  'logit_binary_crossentropy', 'logit_categorical_crossentropy', 'logit_crossentropy',
  'binary_crossentropy', 'categorical_crossentropy'
]

def logit_binary_crossentropy(target, predictions, weights=None):
  assert len(predictions.shape) == 1, 'Predictions for a binary loss must be a 1D-tensor.'
  losses = target * tf.nn.softplus(-predictions) + (1 - target) * tf.nn.softplus(predictions)

  if weights is not None:
    return tf.reduce_mean(weights * losses)
  else:
    return tf.reduce_mean(losses)

def logit_categorical_crossentropy(target, predictions, weights=None):
  assert len(predictions.shape) == 2, 'Predictions for a categorical loss must be a 1D-tensor.'
  normed_predictions = predictions - tf.reduce_max(predictions, axis=1)[:, None]
  neg_log_softmax = tf.math.reduce_logsumexp(normed_predictions, axis=1)[:, None] - normed_predictions

  losses = tf.reduce_sum(target * neg_log_softmax, axis=1)

  if weights is not None:
    return tf.reduce_mean(weights * losses)
  else:
    return tf.reduce_mean(losses)

def logit_crossentropy(target, predictions, weights=None):
  if len(target.shape) == 1:
    return logit_binary_crossentropy(target, predictions, weights)
  else:
    return logit_categorical_crossentropy(target, predictions, weights)

def binary_crossentropy(target, predictions, eps=1e-3, weights=None):
  assert len(predictions.shape) == 1, 'Predictions for binary loss must be a 1D-tensor.'

  losses = target * tf.math.log(predictions + eps) + (1 - target) * tf.math.log(1 - predictions + eps)

  if weights is not None:
    return -tf.reduce_mean(weights * losses)
  else:
    return -tf.reduce_mean(losses)

def categorical_crossentropy(target, predictions, weights=None):
  assert len(predictions.shape) == 2, 'Predictions for a categorical loss must be a 1D-tensor.'

  losses = target * tf.math.log(predictions)
  if weights is not None:
    return -tf.reduce_mean(weights * losses)
  else:
    return -tf.reduce_mean(losses)

def crossentropy(target, predictions, weights=None):
  if len(target.shape) == 1:
    return binary_crossentropy(target, predictions, weights)
  else:
    return categorical_crossentropy(target, predictions, weights)

def _concat(p_neg, p_pos, keep_priors=True):
  predictions = tf.concat([
    p_neg,
    p_pos
  ], axis=0)

  target = tf.concat([
    tf.zeros_like(p_neg),
    tf.ones_like(p_pos),
  ], axis=0)

  if not keep_priors:
    n_neg = tf.shape(p_neg)[0]
    n_pos = tf.shape(p_pos)[1]

    total = n_neg + n_pos

    w_neg = 0.5 * total / n_neg
    w_pos = 0.5 * total / n_pos

    weights = tf.concat([
      w_neg * tf.ones_like(p_neg),
      w_pos * tf.ones_like(p_pos),
    ])
  else:
    weights = None

  return target, predictions, weights

def per_class_logit_crossentropy(predictions_negative, predictions_positive, keep_priors=True):
  target, predictions, weights = _concat(predictions_negative, predictions_positive, keep_priors=keep_priors)

  return logit_binary_crossentropy(target, predictions, weights )



import tensorflow as tf

from ..parameters import glorot_scaling

__all__ = [
  'reg_l1', 'reg_l2', 'reg_elastic_net',
  'norm_preserving_l2'
]

def reg_l2():
  def reg(param):
    return tf.reduce_sum(param() ** 2)

  return reg

def reg_l1():
  def reg(param):
    return tf.reduce_sum(tf.abs(param()))

  return reg

def reg_elastic_net(alpha=0.1):
  def reg(param):
    return (1 - alpha) * reg_l2()(param) + alpha * reg_l1()(param)
  return reg


def norm_preserving_l2(scale_f=glorot_scaling):
  def reg(param):
    scale = scale_f(param.shape)
    norm = tf.sqrt(
      tf.reduce_mean(param() ** 2)
    )

    return (norm - scale) ** 2

  return reg


import numpy as np
import tensorflow as tf

from .. import layers
from ..parameters import FreeParameter

__all__ = [
  'normalize_weights'
]


def normalize_weights(network, batch_f, weights=True, biases=True):
  """
  WARNING: this procedure assumes that output of each layer is sublinear with respect to its inputs, i.e.:
  `f(x, scale * W, b) <= scale * f(x, W, b)`
  and
  `f(x, W, b + offset) <= f(x, W, b) + offset`

  While this assumption is satisfied for the most of the commonly used layers, care should be taken.
  """
  for layer in network.layers():
    if weights:
      weight_parameters = layers.get_parameters(layer, weights=True)
    else:
      weight_parameters = []

    if biases:
      bias_parameters = layers.get_parameters(layer, biases=True)
    else:
      bias_parameters = []

    W_vars = [
      var
      for weight in weight_parameters
      if isinstance(weight, FreeParameter)

      for var in weight.variables()
    ]

    b_vars = [
      var
      for bias in bias_parameters
      if isinstance(bias, FreeParameter)

      for var in bias.variables()
    ]

    if len(b_vars) > 0:
      output = layers.get_output(layer, substitutes=dict(zip(network.inputs, batch_f())))
      mean = np.mean(output)

      offset = mean / len(b_vars)

      for var in b_vars:
        var.assign(var - offset)

    if len(W_vars) > 0:
      output = layers.get_output(layer, substitutes=dict(zip(network.inputs, batch_f())))
      std = np.std(output)

      scale = np.power(1 / std, 1 / len(W_vars))

      for var in W_vars:
        var.assign(var * scale)

def __normalize_output(
  network, batch_f, optimizer=None,
  n_iters=16 * 1024,
  target_mean=0.0, target_variance=1.0,
  strategy=None
):
  augmented_parameters = {
    layer : [
      strategy(param)
      for param in layers.get_parameters(layer)
    ]
    for layer in network.layers()
  }

  augments = [
    augments()
    for layer in augmented_parameters
    for param in augmented_parameters[layer]
    for augments in param.get_augments()
  ]

  def layer_output(layer, *inputs, **modes):
    param_values = [
      param()
      for param in augmented_parameters[layer]
    ]

    return apply_with_kwargs(layer.get_output_for, *param_values, *inputs, **modes)

  get_output = layers.reduce_graph(lambda layer: lambda *inputs, **modes: layer_output(layer, *inputs, **modes), strict=True)

  @tf.function(autograph=False)
  def call():
    args = batch_f()
    result, = get_output(network.outputs, substitutes=dict(zip(network.inputs, args)))
    return result

  @tf.function
  def target():
    result = call()
    mean = tf.reduce_mean(result)
    variance = tf.reduce_mean((result - mean) ** 2)

    return (mean - target_mean) ** 2 + (variance - target_variance) ** 2

  opt = optimizer(target, augments)

  for _ in range(n_iters):
    opt()

  for layer in augmented_parameters:
    for param in augmented_parameters[layer]:
      param.apply()








import tensorflow as tf

from ...meta import derive
from ..meta import FunctionalLayer, model_from

from .conv_utils import *

__all__ = [
  'GeneralPooling1DLayer', 'GeneralPooling2DLayer', 'GeneralPooling3DLayer',
  'MaxPool1DLayer', 'MaxPool2DLayer', 'MaxPool3DLayer',
  'MeanPool1DLayer', 'MeanPool2DLayer', 'MeanPool3DLayer',

  'general_pool_1d', 'general_pool_2d', 'general_pool_3d', 'general_pool',
  'mean_pool_1d', 'mean_pool_2d', 'mean_pool_3d', 'mean_pool',
  'max_pool_1d', 'max_pool_2d', 'max_pool_3d', 'max_pool',

  'GlobalPoolLayer', 'global_pool',
    
  'GlobalMaxPool1D', 'GlobalMaxPool2D', 'GlobalMaxPool3D',
  'GlobalMeanPool1D', 'GlobalMeanPool2D', 'GlobalMeanPool3D',

  'global_max_pool_1d', 'global_max_pool_2d', 'global_max_pool_3d', 'global_max_pool',
  'global_mean_pool_1d', 'global_mean_pool_2d', 'global_mean_pool_3d', 'global_mean_pool'
]

class PoolingLayer(FunctionalLayer):
  def __init__(self, incoming,
               ndim=None,
               kernel_size=2,
               pooling_function='max',
               pad='valid',
               stride=None,
               name=None):
    super(PoolingLayer, self).__init__(incoming, name=name)

    self.ndim = ndim
    self.pooling_op = get_pool_op(pooling_function, ndim)

    self.kernel_size = get_kernel_size(kernel_size, ndim=ndim)

    if stride is None:
      stride = self.kernel_size

    self.stride = get_kernel_size(stride, ndim=ndim)

    self.pad = pad.lower()

    self.pooling_function = pooling_function.lower()

  def get_output_for(self, input, channel_last=False):
    if channel_last:
      input = to_channel_last(input, ndim=self.ndim)

    pooled = self.pooling_op(
      input,
      ksize=self.kernel_size,
      strides=self.stride,
      padding=get_padding(self.pad),
      data_format=get_data_format(self.ndim, channel_last=channel_last),
      name=str(self) + '_pool%dd' % self.ndim,
    )

    if channel_last:
      pooled = to_channel_fist(pooled, self.ndim)

    return pooled

  def get_output_shape_for(self, input_shape):
    if len(input_shape) != self.ndim + 2:
      raise ValueError(
        '%dD pool layer accepts only %dD tensors [got %s]!' % (input_shape, self.ndim, len(input_shape))
      )

    return (input_shape[0], input_shape[1]) + tuple([
      get_conv_output_shape(
        input_shape[i + 2],
        self.kernel_size[i],
        stride=self.stride[i],
        dilation=1,
        pad=self.pad
      )
      for i in range(self.ndim)
    ])


GeneralPooling1DLayer = derive('GeneralPooling1D').based_on(PoolingLayer).with_fixed(ndim=1)
GeneralPooling2DLayer = derive('GeneralPooling2D').based_on(PoolingLayer).with_fixed(ndim=2)
GeneralPooling3DLayer = derive('GeneralPooling3D').based_on(PoolingLayer).with_fixed(ndim=3)

general_pool_1d = model_from(GeneralPooling1DLayer)()
general_pool_2d = model_from(GeneralPooling2DLayer)()
general_pool_3d = model_from(GeneralPooling3DLayer)()

general_pool = select_by_dim([None, general_pool_1d, general_pool_2d, general_pool_3d])

MaxPool1DLayer = derive('MaxPool1DLayer').based_on(PoolingLayer).with_fixed(ndim=1, pooling_function='max', stride=None)
MaxPool2DLayer = derive('MaxPool2DLayer').based_on(PoolingLayer).with_fixed(ndim=2, pooling_function='max', stride=None)
MaxPool3DLayer = derive('MaxPool3DLayer').based_on(PoolingLayer).with_fixed(ndim=3, pooling_function='max', stride=None)

max_pool_1d = model_from(MaxPool1DLayer)()
max_pool_2d = model_from(MaxPool2DLayer)()
max_pool_3d = model_from(MaxPool3DLayer)()

max_pool = select_by_dim([None, max_pool_1d, max_pool_2d, max_pool_3d])

MeanPool1DLayer = derive('MeanPool1DLayer').based_on(PoolingLayer).with_fixed(ndim=1, pooling_function='mean', stride=None)
MeanPool2DLayer = derive('MeanPool2DLayer').based_on(PoolingLayer).with_fixed(ndim=2, pooling_function='mean', stride=None)
MeanPool3DLayer = derive('MeanPool3DLayer').based_on(PoolingLayer).with_fixed(ndim=3, pooling_function='mean', stride=None)

mean_pool_1d = model_from(MeanPool1DLayer)()
mean_pool_2d = model_from(MeanPool2DLayer)()
mean_pool_3d = model_from(MeanPool3DLayer)()

mean_pool = select_by_dim([None, mean_pool_1d, mean_pool_2d, mean_pool_3d])

class GlobalPoolLayer(FunctionalLayer):
  def __init__(self, incoming, pool_f, axis, name=None):
    super(GlobalPoolLayer, self).__init__(incoming, name=name)

    self.pool_f = pool_f
    self.axis = axis

  def get_output_for(self, incoming):
    return self.pool_f(incoming, axis=self.axis)

  def get_output_shape_for(self, incoming_shape):
    return tuple([
      incoming_shape[i] for i in range(len(incoming_shape)) if i not in self.axis
    ])


global_pool = model_from(GlobalPoolLayer)()

GlobalMaxPool1D = derive('GlobalMaxPool1DLayer').based_on(GlobalPoolLayer).with_fixed(pool_f=tf.reduce_max, axis=(2, ))
GlobalMaxPool2D = derive('GlobalMaxPool2DLayer').based_on(GlobalPoolLayer).with_fixed(pool_f=tf.reduce_max, axis=(2, 3))
GlobalMaxPool3D = derive('GlobalMaxPool3DLayer').based_on(GlobalPoolLayer).with_fixed(pool_f=tf.reduce_max, axis=(2, 3, 4))

global_max_pool_1d = model_from(GlobalMaxPool1D)()
global_max_pool_2d = model_from(GlobalMaxPool2D)()
global_max_pool_3d = model_from(GlobalMaxPool3D)()

global_max_pool = select_by_dim([None, global_max_pool_1d, global_max_pool_2d, global_max_pool_3d])

GlobalMeanPool1D = derive('GlobalMeanPool1DLayer').based_on(GlobalPoolLayer).with_fixed(pool_f=tf.reduce_mean, axis=(2, ))
GlobalMeanPool2D = derive('GlobalMeanPool2DLayer').based_on(GlobalPoolLayer).with_fixed(pool_f=tf.reduce_mean, axis=(2, 3))
GlobalMeanPool3D = derive('GlobalMeanPool3DLayer').based_on(GlobalPoolLayer).with_fixed(pool_f=tf.reduce_mean, axis=(2, 3, 4))

global_mean_pool_1d = model_from(GlobalMeanPool1D)()
global_mean_pool_2d = model_from(GlobalMeanPool2D)()
global_mean_pool_3d = model_from(GlobalMeanPool3D)()

global_mean_pool = select_by_dim([None, global_mean_pool_1d, global_mean_pool_2d, global_mean_pool_3d])


import tensorflow as tf

__all__ = [
  'normalize_axis'
]

def normalize_axis(tensor_or_dim, axis):
  if isinstance(tensor_or_dim, int):
    dim = tensor_or_dim
  else:
    dim = len(tensor_or_dim.shape)

  if isinstance(axis, int):
    return (axis % dim, )
  else:
    return tuple(a % dim for a in axis)


import tensorflow as tf

from ..reducers import moving_sum, moving_average, max_average

__all__ = [
  'no_scaling',
  'l2_scaling',
  'total_path_scaling',
  'max_scaling'
]

class NoScaling(object):
  def __init__(self, var_spec):
    pass

  def __call__(self, grads):
    return [None for _ in grads]

def no_scaling():
  def scaling(var_spec):
    return NoScaling(var_spec)

  return scaling

class TotalPathScaling(object):
  def __init__(self, var_spec):
    self.moments = moving_sum(*var_spec)

  def __call__(self, grads):
    return [
      tf.sqrt(m)
      for m in self.moments([ g ** 2 for g in grads ])
    ]

def total_path_scaling():
  def scaling(var_spec):
    return TotalPathScaling(var_spec)

  return scaling

class L2Scaling(object):
  def __init__(self, var_spec, rho=0.99, eps=1e-7):
    self.moments = moving_average(*var_spec, rho=rho)
    self.eps = tf.constant(eps, dtype=tf.float32)

  def __call__(self, grads):
    return [
      tf.sqrt(m) + self.eps
      for m in self.moments([ g ** 2 for g in grads ])
    ]

def l2_scaling(rho=0.99, eps=1e-7):
  def scaling(var_spec, ):
    return L2Scaling(var_spec, rho=rho, eps=eps)

  return scaling


class MaxScaling(object):
  def __init__(self, var_spec, rho=0.99, eps=1e-7):
    self.moments = max_average(var_spec, rho=rho)
    self.eps = tf.constant(eps, dtype=tf.float32)

  def __call__(self, grads):
    return [
      m + self.eps
      for m in self.moments([ tf.abs(g) for g in grads ])
    ]

def max_scaling(rho=0.99, eps=1e-7):
  def scaling(var_spec, ):
    return MaxScaling(var_spec, rho=rho)

  return scaling


import numpy as np
import tensorflow as tf

from .meta import *
from .parameter_utils import *

__all__ = [
  'zeros_init',
  'ones_init',
  'const_init',
  'normal_init',
  'uniform_init',
  'glorot_normal_init',
  'glorot_uniform_init',

  'glorot_scaling',

  'glorot_normal_double_init',

  'default_weight_init',
  'default_bias_init',

  'shared_parameter'
]

zeros_init = as_free_parameter(tf.zeros)
ones_init = as_free_parameter(tf.ones)


def _const_init(shape, value, dtype='float32', name=None):
  if hasattr(value, 'shape') and len(value.shape) > 0:
    assert value.shape == shape, \
      'If `value` is a non-scalar array then `shape` (%s) must be equal to `value.shape` (%s)' % (shape, value.shape)

    return tf.constant(value, dtype=dtype, name=init_name(name))
  else:
    return tf.ones(shape=shape, dtype=dtype, name=init_name(name)) * \
           tf.constant(value, dtype=dtype)

const_init = as_free_parameter(_const_init)

normal_init = as_free_parameter(tf.random.normal)

uniform_init = as_free_parameter(tf.random.uniform)


def glorot_scaling(shape, gain=1.0):
  in_units, out_units = shape[-2:]
  receptive_field_area = np.prod(shape[:-2])

  return gain * np.sqrt(2.0 / ((in_units + out_units) * receptive_field_area))


def _glorot_normal_init(shape, gain=1.0, dtype='float32', name=None):
  if len(shape) < 2:
    return tf.random.normal(shape=shape, mean=0.0, stddev=gain, dtype=dtype, name=init_name(name))
  else:
    scale = glorot_scaling(shape, gain)
    return tf.random.normal(shape=shape, mean=0.0, stddev=scale, dtype=dtype, name=init_name(name))

def _double_glorot_normal_init(shape, target_shape, gain=1.0, dtype='float32', name=None):
  gain = glorot_scaling(target_shape, gain=gain)
  return _glorot_normal_init(shape, gain=gain, dtype=dtype, name=name)


glorot_normal_init = as_free_parameter(_glorot_normal_init)
glorot_normal_double_init = as_free_parameter(_double_glorot_normal_init)


def _glorot_uniform_init(shape, gain=1.0, dtype='float32', name=None):
  if len(shape) < 2:
    return tf.random.normal(shape=shape, mean=0.0, stddev=gain, dtype=dtype, name=init_name(name))
  else:
    scale = tf.constant(
      glorot_scaling(shape, gain) * np.sqrt(3),
      dtype=dtype
    )

    return tf.random.uniform(shape=shape, minval=-scale, maxval=scale, dtype=dtype, name=name)

glorot_uniform_init = as_free_parameter(_glorot_uniform_init)


default_weight_init = glorot_normal_init(gain=1.0)
default_bias_init = zeros_init()


class SharedParameter(Parameter):
  def __init__(self, parent : Parameter, properties, name=None):
    self.parent = parent
    
    super(SharedParameter, self).__init__(
      self.parent.shape,
      properties=properties,
      name=name
    )

  def dependencies(self):
    return self.parent.dependencies()

  def variables(self):
    ### shared parameters do not own variables, so
    ### this method outputs empty list.
    return []

  def get_output_for(self, *vars):
    return self.parent.get_output_for(*vars)

  def reset(self):
    pass

  def __str__(self):
    return 'SharedParameter[parent=%s]' % (self.parent, )


class ParameterCloneMachine(object):
  def __init__(self, parameter_constructor):
    self.parameter_constructor = parameter_constructor
    self.parameter = None

  def __call__(self, shape, name=None, **additional_properties):
    if self.parameter is None:
      self.parameter = self.parameter_constructor(shape, name, **additional_properties)
      return self.parameter
    else:
      assert shape == self.parameter.shape, 'Can not clone parameter for different shape.'

      return SharedParameter(
        self.parameter,
        properties=combine_properties(
          self.parameter_constructor.user_defined_properties(),
          additional_properties
        ),
        name=name
      )

shared_parameter = ParameterCloneMachine


import tensorflow as tf

__all__ = [
  'sigmoid',
  'leaky_sigmoid',
  'tanh',
  'leaky_tanh',
  'relu',
  'leaky_relu',
  'softplus',
  'softmax',
  'elu',
  'linear',

  'gaussian',

  'default_semibounded_nonlinearity',
  'default_bounded_nonlinearity',
  'default_nonlinearity'
]

class Nonlinearity(object):
  def __init__(self, nonlinearity, hyperparameters):
    self.nonlinearity = nonlinearity
    self.hyperparameters = hyperparameters

  def __call__(self, x):
    return self.nonlinearity(x)

  def __str__(self):
    return '%s(%s)' % (
      self.__class__.__name__,
      ', '.join([
        '%s=%s' % (k, v) for k, v in self.hyperparameters.items()
      ])
    )

def nonlinearity_from(f):
  import inspect

  signature = inspect.signature(f)
  name_parameter, = [
    signature.parameters[p] for p in signature.parameters if p == 'name'
  ]

  name = name_parameter.default

  def __init__(self, *args, **kwargs):
    bound = signature.bind(*args, **kwargs)
    bound.apply_defaults()

    g = f(*bound.args, **bound.kwargs)

    hyperparameters = bound.arguments.copy()
    hyperparameters.pop('name', None)

    Nonlinearity.__init__(self, g, hyperparameters)

  clazz = type(
    name,
    (Nonlinearity, ),
    dict(__init__ = __init__)
  )

  return clazz

sigmoid = nonlinearity_from(
  lambda name='sigmoid': lambda x: tf.nn.sigmoid(x, name=name)
)

leaky_sigmoid = nonlinearity_from(
  lambda leakiness=0.05, name='leaky_sigmoid': lambda x: \
    tf.nn.sigmoid(x, name=name) + leakiness * x
)

tanh = nonlinearity_from(
  lambda name='tanh': lambda x: tf.tanh(x, name=name)
)

leaky_tanh = nonlinearity_from(
  lambda leakiness=0.05, name='leaky_tanh': lambda x: \
    tf.tanh(x, name=name) + leakiness * x
)

relu = nonlinearity_from(
  lambda name='ReLU': lambda x: tf.nn.relu(x, name=name)
)

leaky_relu = nonlinearity_from(
  lambda leakiness=0.05, name='leaky_ReLU': \
    lambda x: tf.nn.leaky_relu(x, alpha=leakiness, name=name)
)

softplus = nonlinearity_from(
  lambda name='softplus': lambda x: tf.nn.softplus(x)
)
softmax = nonlinearity_from(
  lambda name='softmax': lambda x: tf.nn.softmax(x, name=name)
)

elu = nonlinearity_from(
  lambda name='ELU': lambda x: tf.nn.elu(x, name=name)
)

linear = nonlinearity_from(
  lambda name='linear': lambda x: x
)

gaussian = nonlinearity_from(
  lambda name='gaussian': lambda x: tf.exp(-x ** 2, name=name)
)

default_bounded_nonlinearity = sigmoid()

### well, leaky relu is not exactly semi-bounded...
default_semibounded_nonlinearity = leaky_relu(0.05)

default_nonlinearity = default_semibounded_nonlinearity


import tensorflow as tf

from .parameter_utils import combine_properties, init_name
from ..meta import bind

__all__ = [
  'Parameter',
  'ConstantParameter',
  'FreeParameter',
  'UnboundParameter', 'BoundParameter',

  'constant_parameter',
  'free_parameter',
  'unbound_parameter',
  'bound_parameter',

  'as_free_parameter',
  'parameter_model'
]

class Parameter(object):
  def __init__(self, shape, properties, name=None):
    self.shape = shape
    self.name = name

    if properties is None:
      self.properties = dict()
    else:
      self.properties = properties

  def __call__(self):
    return self.get_output_for(*self.dependencies())

  def get_output_for(self, *vars):
    raise NotImplementedError()

  def dependencies(self):
    """
    Returns variables this parameter depends on, i.e. ones
    required for `get_output_for` method.
    `dependencies()` can be different from `variables()` if `Parameter`
    does not own variables, like a shared parameter.

    :return: list of `tf.Variable`
    """
    return self.variables()

  def variables(self):
    """
    Returns variables that are owned by the parameter.
    A variable shared between multiple parameters must be owned by only one of the parameters,
    however, might appear in `dependencies()` multiple times.

    :return: list of `tf.Variable`
    """
    raise NotImplementedError()

  def reset(self):
    raise NotImplementedError()

  def __str__(self):
    name = self.__class__.__name__ if self.name is None else self.name
    shape = 'x'.join([ '%d' % (s, ) for s in self.shape ])
    props = [('shape', shape)] + list(self.properties.items())

    return '%s (%s)' % (
      name,
      ', '.join([ '%s=%s' % (k, v) for k, v in props ])
    )

  def __repr__(self):
    return str(self)

class ConstantParameter(Parameter):
  def __init__(self, value, properties=None, name=None):
    super(ConstantParameter, self).__init__(value.shape, properties, name=name)

    self._original_value = value
    self._value = tf.constant(value, dtype=value.dtype, shape=value.shape, name=name)

  def __call__(self):
    return self._value

  def get_output_for(self,):
    return self._value

  def variables(self):
    return []

  def reset(self):
    return self._value


def constant_parameter(value, dtype=tf.float32, name=None, **properties):
  def constructor(shape, **additional_properties):
    import numpy as np

    try:
      dtype_ = dtype.as_numpy_dtype
    except AttributeError:
      dtype_ = dtype

    if not isinstance(value, np.ndarray):
      v = np.array(value, dtype=dtype_)
    else:
      v = value

    v = np.broadcast_to(v, shape)
    props = combine_properties(properties, additional_properties)

    return ConstantParameter(v, properties=props, name=name)

  return constructor


class FreeParameter(Parameter):
  ### everything for pickle
  def __init__(self, initializer, initializer_arguments, properties=None, name=None):
    super(FreeParameter, self).__init__(initializer_arguments['shape'], properties, name)

    self._initializer = initializer
    self._initializer_arguments = initializer_arguments

    self.initial = initializer(**initializer_arguments)
    self._value = tf.Variable(
      initial_value=self.initial,
      name=name,
      dtype=self.initial.dtype,
      trainable=self.properties.get('trainable', False)
    )

  def get_output_for(self, value):
    return value

  def variables(self):
    return [self._value,]

  def reset(self):
    return self._value.assign(
      self._initializer(**self._initializer_arguments)
    )

free_parameter = FreeParameter


def parameter_model2(clazz):
  from inspect import signature, Signature, Parameter
  from ..meta import get_kwargs

  original_signature = signature(clazz)

  assert any([
    param.name == 'shape'
    for param in original_signature.parameters.values()
  ]), 'Decorated function or class must have a "shape" argument.'

  accepted_kwargs = get_kwargs(clazz)

  def wrapped(name=None, **kwargs):
    external_name = name

    def constructor(shape, name=None, **additional_properties):
      if external_name is not None:
        name = external_name

      if name is not None:
        name = init_name(name)

      constructor_kwargs = dict([
        (k, v) for k, v in kwargs.items() if k in accepted_kwargs
      ])

      if 'name' in accepted_kwargs:
        constructor_kwargs['name'] = name

      properties = dict([
        (k, v) for k, v in kwargs.items() if k not in accepted_kwargs
      ])
      properties = combine_properties(properties, additional_properties)

      clazz_args, clazz_kwargs = bind(original_signature, shape=shape, **constructor_kwargs, properties=properties)

      return clazz(*clazz_args, **clazz_kwargs)

    return constructor

  new_signature = Signature(
    parameters=[
        param for param in original_signature.parameters.values() if param.name != 'shape'
      ] + ([] if 'name' in accepted_kwargs else [
        Parameter('name', kind=Parameter.POSITIONAL_OR_KEYWORD, annotation='str')
      ]),
    return_annotation=original_signature.return_annotation
  )

  wrapped.__signature__ = new_signature
  wrapped.__doc__ = clazz.__doc__

  return wrapped

class ParameterConstructor(object):
  def __init__(self, user_defined_properties, external_name=None):
    self._user_defined_properties = user_defined_properties
    self._external_name = external_name

  def user_defined_properties(self):
    return self._user_defined_properties

  def __call__(self, shape, name=None, **additional_properties):
    raise NotImplementedError

def parameter_model(clazz):
  from inspect import signature, Signature, Parameter as FParameter
  from ..meta import get_kwargs

  original_signature = signature(clazz)

  assert any([
    param.name == 'shape'
    for param in original_signature.parameters.values()
  ]), 'Decorated function or class must have a "shape" argument.'

  assert any([
    param.name == 'properties'
    for param in original_signature.parameters.values()
  ]), 'Decorated function or class must have a "properties" argument.'

  accepted_kwargs = get_kwargs(clazz)

  def __init__(self, name=None, **kwargs):
    user_defined_properties = dict([
      (k, v) for k, v in kwargs.items()
      if k not in accepted_kwargs
    ])

    self.constructor_kwargs = dict([
      (k, v) for k, v in kwargs.items()
      if k in accepted_kwargs
    ])

    ParameterConstructor.__init__(self, user_defined_properties, external_name=name)

  new_signature = Signature(
    parameters=[
       FParameter('self', kind=FParameter.POSITIONAL_ONLY)
     ] + [
       param for param in original_signature.parameters.values() if param.name != 'shape'
     ] + ([] if 'name' in accepted_kwargs else [
      FParameter('name', kind=FParameter.POSITIONAL_OR_KEYWORD, annotation='str')
    ]),
    return_annotation=original_signature.return_annotation
  )

  __init__.__signature__ = new_signature
  __init__.__doc__ = clazz.__doc__

  def __call__(self, shape, name=None, **additional_properties):
    if self._external_name is not None:
      name = self._external_name

    if name is not None:
      name = init_name(name)

    constructor_kwargs = self.constructor_kwargs.copy()

    if 'name' in accepted_kwargs:
      constructor_kwargs['name'] = name

    properties = combine_properties(self._user_defined_properties, additional_properties)

    clazz_args, clazz_kwargs = bind(original_signature, shape=shape, **constructor_kwargs, properties=properties)

    return clazz(*clazz_args, **clazz_kwargs)

  return type(
    clazz.__name__,
    (ParameterConstructor, ),
    dict(
      __init__=__init__,
      __call__=__call__
    )
  )

def as_free_parameter(f):
  from inspect import signature, Signature, Parameter as FParameter
  from ..meta import get_kwargs

  f_signature = signature(f)

  assert not any([
    param.kind == FParameter.VAR_KEYWORD
    for param in f_signature.parameters.values()
  ]), 'parameter decorator does not accept functions with `**kwargs`.'

  assert any([
    param.name == 'shape'
    for param in f_signature.parameters.values()
  ]), 'Decorated function must have a "shape" argument.'

  accepted_kwargs = get_kwargs(f)

  def __init__(self, *args, name=None, **kwargs):
    self.external_name = name
    self.args = args
    self.kwargs = kwargs

    user_defined_properties = dict([
      (k, v) for k, v in kwargs.items() if k not in accepted_kwargs
    ])

    self.function_kwargs = dict([
      (k, v) for k, v in kwargs.items() if k in accepted_kwargs
    ])

    ParameterConstructor.__init__(self, user_defined_properties, external_name=name)

  new_signature = Signature(
    parameters=[
      FParameter('self', kind=FParameter.POSITIONAL_ONLY)
    ] + [
      param for param in f_signature.parameters.values() if param.name != 'shape'
    ] + ([] if 'name' in accepted_kwargs else [
      FParameter('name', kind=FParameter.POSITIONAL_OR_KEYWORD, annotation='str')
    ]) + [
      FParameter('properties', kind=FParameter.VAR_KEYWORD)
    ],
    return_annotation=f_signature.return_annotation
  )

  __init__.__signature__ = new_signature
  __init__.__doc__ = f.__doc__

  def __call__(self, shape, name=None, **additional_properties):
    if self._external_name is not None:
      name = self._external_name

    if name is not None:
      name = init_name(name)

    function_kwargs = self.function_kwargs.copy()

    if 'name' in accepted_kwargs:
      function_kwargs['name'] = name

    arguments = f_signature.bind(*self.args, shape=shape, **function_kwargs)
    arguments.apply_defaults()

    return FreeParameter(
      initializer=f, initializer_arguments=arguments.arguments,
      name=name, properties=combine_properties(self._user_defined_properties, additional_properties)
    )

  clazz = type(
    f.__name__,
    (ParameterConstructor, ),
    dict(
      __init__=__init__,
      __call__=__call__
    )
  )

  return clazz

class BoundParameter(Parameter):
  def __init__(self, value, shape=None, properties=None, name=None):
    super(BoundParameter, self).__init__(shape, properties, name)

    self.value = value
    self.shape = tuple(value.shape.as_list()) if shape is None else shape
    self.dtype = value.dtype

  def get_output_for(self,):
    return self.value

  def variables(self):
    return []

  def __str__(self):
    return '%s [%s]: %s' % (
      self.__class__.__name__ if self.name is None else self.name,
      self.value,
      str(self.shape),
    )

bound_parameter = lambda value, shape=None, name=None, **properties: \
  BoundParameter(value, shape, properties=properties, name=name)

class UnboundParameter(Parameter):
  def __init__(self, shape, dtype='float32', properties=None, name=None):
    super(UnboundParameter, self).__init__(shape, properties, name)

    self.shape = shape
    self.dtype = dtype

  def get_output_for(self, ):
    raise NotImplementedError('Unbound parameters are meant to be substituted.')

  def own_variables(self):
    raise NotImplementedError('Unbound parameters are meant to be substituted.')

  def __str__(self):
    return '%s: %s' % (
      self.__class__.__name__ if self.name is None else self.name,
      str(self.shape)
    )

unbound_parameter = parameter_model(UnboundParameter)


### With a little shame stolen from nolearn

"""
Copyright (c) 2012-2015 Daniel Nouri

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from ..layers import Layer, DenseLayer, InputLayer, get_output_shape, get_layers
from ..layers.inspect import get_number_of_parameters, get_total_number_of_parameters
from ..networks import Network

__all__ = [
  'draw_to_file',
  'draw_to_notebook',
  'viz_all_params',
  'viz_params'
]

_color_set = [
  ### blue
  ('#a6cee3', '#1f78b4'),
  ### green
  ('#b2df8a', '#33a02c'),
  ### red
  ('#fb9a99', '#e31a1c'),
  ### orange
  ('#fdbf6f', '#ff7f00'),
  ### magenta
  ('#cab2d6', '#6a3d9a'),
  ### yellow/brown
  ('#ffff99', '#b15928'),
]

def _stable_hash(data):
  from hashlib import blake2b
  return int.from_bytes(blake2b(data.encode('utf-16be'), digest_size=8).digest(), byteorder='big')

def get_color(layer_class):
  layer_type = layer_class.__name__.lower()

  hashed = _stable_hash(layer_class.__name__) % len(_color_set[0])

  if 'conv' in layer_type:
    return _color_set[0][hashed]

  if issubclass(layer_class, DenseLayer) or 'dense' in layer_type:
    return _color_set[1][hashed]

  if issubclass(layer_class, InputLayer) or 'input' is layer_type:
    return _color_set[2][hashed]

  if 'pool' in layer_type:
    return _color_set[3][hashed]

  if 'recurrent' in layer_type:
    return _color_set[4][hashed]

  return _color_set[5][hashed]

def viz_params(**kwargs):
  def f(layer):
    param_info = get_number_of_parameters(layer, **kwargs).items()
    if len(param_info) > 0:
      return ','.join([ '%s: %d' % (k.name, v) for k, v in param_info ])
    else:
      return None

def viz_all_params(**properties):
  def number_of_params(layer):
    n = get_number_of_parameters(layer, **properties)
    if n > 0:
      return '%d' % n
    else:
      return None

  return number_of_params

def viz_name(force_class_name=False, remove_layer_from_class_name=True, kernel_info=True):
  def name(layer):
    layer_class = layer.__class__.__name__.split('.')[-1]

    if remove_layer_from_class_name:
      layer_class = layer_class.replace('Layer', '')
      layer_class = layer_class.replace('layer', '')

    kernel_info = []
    if hasattr(layer, 'kernel_size'):
      kernel_info.append(
        'x'.join('%d' % k for k in layer.kernel_size)
      )

    if hasattr(layer, 'stride'):
      if any(s != 1 for s in layer.stride):
        kernel_info.append(
          'stride=%s' % (
            'x'.join('%d' % k for k in layer.stride),
          )
        )

    if hasattr(layer, 'padding'):
      if layer.padding != 'valid':
        kernel_info.append('pad=%s' % layer.padding)

    if layer.name is None or force_class_name:
      result = layer_class
    else:
      result = '%s : %s' % (layer.name, layer_class)

    if kernel_info:
      return '%s %s' % (result, ', '.join(kernel_info))
    else:
      return result

  return name

default_display_properties = [
  (None, viz_name()),
  ('#params', viz_all_params()),
  ('output shape', get_output_shape),
]

def make_graph(layers, properties_to_display=default_display_properties):
  import pydotplus as pydot

  graph = pydot.Dot('network', graph_type='digraph')

  layer_indx = dict([
    (layer, 'node%d' % i) for i, layer in enumerate(layers)
  ])

  nodes = dict()

  properties_spec = []

  for spec in properties_to_display:
    try:
      name, prop = spec
      properties_spec.append((name, prop))
    except ValueError:
      properties_spec.append((None, spec))

  for layer in layers:
    info = []
    for prop_name, prop in properties_spec:
      try:
        result = prop(layer)
        if result is None:
          continue

        if prop_name is not None:
          info.append('%s: %s' % (prop_name, result))
        else:
          info.append('%s' % result)
      except Exception as e:
        import warnings
        warnings.warn('Failed to evaluate property %s [%s]' % (prop_name, e))

    nodes[layer] = pydot.Node(
      name=layer_indx[layer], shape='record',
      label='\n'.join(info),
      fillcolor=get_color(layer.__class__), style='filled'
    )

  for node in nodes.values():
    graph.add_node(node)

  for layer in layers:
    for incoming in getattr(layer, 'incoming', list()):
      graph.add_edge(pydot.Edge(nodes[incoming], nodes[layer]))

  graph.set('splines', 'ortho')

  return graph


def get_png(layers_or_layer_or_network, properties_to_display=default_display_properties):
  import pydotplus as pydot

  try:
    if isinstance(layers_or_layer_or_network, Network):
      layers_or_layer = layers_or_layer_or_network.layers()
    else:
      layers_or_layer = layers_or_layer_or_network

    graph = make_graph(get_layers(layers_or_layer), properties_to_display)

    return graph.create(format='png')
  except pydot.InvocationException as e:
    import traceback
    tb = traceback.format_exc()

    import warnings
    warnings.warn(tb)

def draw_to_file(layers_or_layer_or_network, path, properties_to_display=default_display_properties):
  png = get_png(layers_or_layer_or_network, properties_to_display=properties_to_display)

  with open(path, 'wb') as f:
    f.write(png)


def draw_to_notebook(layers_or_layer_or_network, properties_to_display=default_display_properties):
  from IPython.display import Image

  png = get_png(layers_or_layer_or_network, properties_to_display=properties_to_display)

  return Image(png)



import tensorflow as tf

from .. import layers
from ..subnetworks import achain

from .network_utils import *

__all__ = [
  'Network', 'network',
  'modify_network'
]

class Network(object):
  def __init__(self, inputs, outputs, **modes):
    self.inputs = inputs
    self.outputs = outputs
    self.modes = modes

    try:
      self.__call__.__signature__ = get_signature(inputs)
    except:
      pass

    named_layers = dict()
    for layer in self.layers():
      if layer.name is None:
        continue

      if layer.name not in named_layers:
        named_layers[layer.name] = layer
      else:
        if isinstance(layer, layers.InputLayer) and isinstance(named_layers[layer.name], layers.InputLayer):
          raise Exception('Collision in input names: %s' % layer.name)

    self._named_layers = named_layers

    self._mode_cache = dict()

  def subnet(self, inputs=None, outputs=None):
    if inputs is None:
      inputs = self.inputs
    else:
      inputs = [
        input if isinstance(input, layers.Layer) else self._named_layers[input]
        for input in inputs
      ]

    if outputs is None:
      outputs = self.outputs
    else:
      outputs = [
        output if isinstance(output, layers.Layer) else self._named_layers[output]
        for output in outputs
      ]

    return Network(inputs, outputs)

  def _as_subnet(self, *incoming):
    from ..subnetworks import subnetwork
    return subnetwork(self.inputs, self.outputs)(*incoming)

  def _map_inputs(self, args, kwargs):
    substitutes = dict(zip(self.inputs, args))

    for name, value in kwargs.items():
      if name in self._named_layers:
        layer = self._named_layers[name]
      else:
        raise Exception('There is no layer with name %s' % (name,))

      if layer in substitutes:
        raise Exception('Value for layer %s is provided twice, via a positional and a keyword arguments' % (name,))

      substitutes[layer] = value

    return substitutes

  @tf.function(autograph=False)
  def __call__(self, *args, **kwargs):
    is_arg_layer = [isinstance(arg, layers.Layer) for arg in args]

    if all(is_arg_layer) and len(is_arg_layer) > 0:
      if len(kwargs) != 0:
        raise Exception('Network as a SubnetworkLayer does not accept kwargs')
      return self._as_subnet(*args)

    if any(is_arg_layer) or any([ isinstance(arg, layers.Layer) for arg in kwargs.values() ]):
      raise NotImplementedError('Network can not be called on a mixture of layers and tensors yet.')

    substitutes = self._map_inputs(args, kwargs)

    try:
      return layers.get_output(self.outputs, substitutes=substitutes, **self.modes)

    except Exception as e:
      inputs_wo_substitute = [
        layer
        for layer in self.inputs
        if layer not in substitutes
      ]

      if len(inputs_wo_substitute) > 0:
        raise ValueError('Not all inputs were provided value, this might be the cause of the error.') from e
      else:
        raise


  def mode(self, **modes):
    if len(modes) == 0:
      return self

    new_modes = self.modes.copy()
    for k, v in modes.items():
      new_modes[k] = v

    mode_key = tuple(new_modes.items())
    if mode_key not in self._mode_cache:
      self._mode_cache[mode_key] = Network(self.inputs, self.outputs, **new_modes)

    self._mode_cache[mode_key]._mode_cache = self._mode_cache

    return self._mode_cache[mode_key]


  def reset(self):
    return [
      param.reset()
      for param in self.parameters()
    ]

  def parameters(self, **properties):
    return layers.get_all_parameters(self.outputs, **properties)

  def variables(self, trainable=None, **properties):
    return layers.get_all_variables(self.outputs, trainable=trainable, **properties)

  def assign(self, variable_values):
    for var, value in zip(self.variables(), variable_values):
      var.assign(value)

  def description(self, short=True, **attributes):
    from ..layers.inspect import graph_description
    return graph_description(self.outputs, short=short, inputs=self.inputs, **attributes)

  def __str__(self):
    return self.description(short=True)

  def __repr__(self):
    return self.description(short=True)

  def total_number_of_parameters(self, **properties):
    from ..layers.inspect import get_total_number_of_parameters
    return get_total_number_of_parameters(self.outputs, **properties)

  def layers(self):
    return layers.get_layers(self.outputs)

  def input_shapes(self):
    return layers.get_output_shape(self.inputs)

  def output_shapes(self):
    return layers.get_output_shape(self.outputs)

  def reg_l2(self, weights=True, **properties):
    from ..regularization import reg_l2
    reg = reg_l2()
    return sum([
      reg(param)
      for param in self.parameters(weights=weights, **properties)
    ])


def __is_shape(shape_or_layer):
  return hasattr(shape_or_layer, '__iter__') and all([ (type(s) is int or s is None) for s in shape_or_layer ])

def _get_input_layer(shape_or_layer, name=None, index=None):
  if __is_shape(shape_or_layer) :
    shape = shape_or_layer

    if name is not None:
      return layers.InputLayer(shape=shape, name=name)
    elif index is not None:
      return layers.InputLayer(shape=shape, name='input%d' % index)
    else:
      return layers.InputLayer(shape=shape, name='input')

  elif isinstance(shape_or_layer, layers.Layer):
    return shape_or_layer


def _make_network(factory, inputs, named_inputs):
  input_layers = []

  for i, input in enumerate(inputs):
    input_layers.append(_get_input_layer(input, name=None, index=i))

  for i, (name, input) in enumerate(named_inputs.items()):
    input_layers.append(_get_input_layer(input, name=name, index=i))

  explicit_names = [ layer.name for layer in input_layers if layer.name is not None ]
  assert len(set(explicit_names)) == len(explicit_names)

  outputs = factory(*input_layers)

  if isinstance(outputs, layers.Layer):
    outputs = [outputs]

  return Network(input_layers, outputs)


network = lambda *inputs, **named_inputs: lambda *factory: \
  _make_network(achain(*factory), inputs, named_inputs)

network.__doc__ = \
"""
Allows nice syntax:
```
  net(session)(input1, input2, ..., named_input=)(
    constructor
  )
```

or

```
  net(session)(input)(
    constructor
  )
```

for single input.
"""

def modify_network(operator, nn : Network):
  """
  Performs `craynn.layers.meta.modify_graph` of network's graph...
  See `craynn.layers.meta.modify_graph` documentation for details.
  :param operator: modification operator.
  :param nn: an instance of Network.
  :return: modified network.
  """
  modified = layers.modify_graph(operator, nn.outputs)
  return nn.__class__(nn.inputs, modified, *nn._args, **nn._kwargs)


import tensorflow as tf

__all__ = [
  'Stateful', 'stateful', 'as_stateful'
]

def _populate(obj, property, value):
  setattr(obj, '_' + property, value)
  setattr(obj, property, lambda : getattr(obj, '_' + property))

class Stateful(object):
  def __init__(self, state=None, updates=None, observables=None):
    self._state = state if state is not None else []
    self._updates = updates if updates is not None else tf.no_op()

    if observables is not None:
      for k, v in observables.items():
        _populate(self, k, v)

    self._observables = observables if observables is not None else dict()

    self._reset_op = tf.group([
      var.initializer
      for var in self._state
    ])

  def updates(self):
    return self._updates

  def reset(self):
    return self._reset_op

  def observables(self):
    return self._observables

stateful = Stateful

def as_stateful(f):
  import inspect

  def new_init(self, state=None, updates=None, observables=None):
    Stateful.__init__(self, state, updates, observables)

  clazz = type(f.__name__, (Stateful, ), dict(
    __init__ = new_init
  ))

  def wrapper(*args, **kwargs):
    g = f(*args, **kwargs)

    def new_g(*args, **kwargs):
      result = g(*args, **kwargs)

      if isinstance(result, Stateful):
        return result
      else:
        state, updates, observables = result
        return clazz(state, updates, observables)

    new_g.__signature__ = inspect.signature(g)

    return new_g

  wrapper.__signature__ = inspect.signature(f)
  return wrapper



import numpy as np

import tensorflow as tf

from ...meta import derive
from ...nonlinearities import default_nonlinearity
from ...parameters import default_weight_init, default_bias_init

from ..common import FunctionalLayer
from ..meta import get_output_shape, model_from

from .conv_utils import *

__all__ = [
  'TransposedConv2DLayer',

  'transposed_conv_2d',
  'deconv_2d',
  'transposed_conv',
  'deconv',

  'ptransposed_conv_2d',
  'pdeconv_2d',
  'ptransposed_conv',
  'pdeconv',
]

class TransposedConvLayer(FunctionalLayer):
  def __init__(self, incoming, ndim,
               num_filters,
               kernel_size=3,
               activation=default_nonlinearity,
               pad='valid', stride=1,
               W=default_weight_init,
               b=default_bias_init,
               name=None):
    super(TransposedConvLayer, self).__init__(incoming, name=name)

    self.ndim = ndim
    self.op = get_deconv_op(ndim)

    input_shape = get_output_shape(incoming)

    if len(input_shape) != self.ndim + 2:
      raise ValueError(
        '%dD deconv layer accepts only %dD tensors [got %s]!' % (self.ndim, self.ndim + 2, input_shape)
      )

    if num_filters <= 0:
      raise ValueError(
        '`num` channels must be > 0!'
      )

    self.input_channels = get_channel_dim(input_shape)
    self.input_size = get_spatial_dims(input_shape)
    self.num_channels = num_filters

    self.kernel_size = get_kernel_size(kernel_size, ndim=ndim)
    self.kernel_shape = self.kernel_size + (self.num_channels, self.input_channels)

    self.stride = get_kernel_size(stride, ndim=ndim)

    self.pad = pad.lower()

    self.parameters = [
      W(self.kernel_shape, name='W', weights=True, conv_kernel=True, trainable=True),
      b((self.num_channels,), name='b', biases=True, trainable=True)
    ]

    self.activation = activation

  def get_output_for(self, W, b, input, channel_last=False):
    if channel_last:
      input = to_channel_last(input, ndim=self.ndim)

    ### apparently, tf.shape(x) and X.shape give different results...
    output_shape = self.get_output_shape_for(input.shape)

    convolved = self.op(
      input, W,
      output_shape=output_shape,
      strides=self.stride,
      padding=get_padding(self.pad),
      data_format=get_data_format(self.ndim, channel_last=channel_last),
      name=str(self) + 'deconv%dd' % self.ndim,
    )

    if channel_last:
      convolved = to_channel_fist(convolved, self.ndim)

    broadcast = (None,) + (slice(None),) + (None,) * self.ndim
    return self.activation(
      convolved + b[broadcast]
    )

  def get_output_shape_for(self, input_shape):
    if len(input_shape) != self.ndim + 2:
      raise ValueError(
        '%dD deconv layer accepts only %dD tensors [got %s]!' % (self.ndim, self.ndim + 2, input_shape)
      )

    return tuple([input_shape[0], self.num_channels] + [
      get_deconv_output_shape(
        input_shape[i + 2],
        kernel_size=self.kernel_size[i],
        stride=self.stride[i],
        pad=self.pad
      )
      for i in range(self.ndim)
    ])


TransposedConv1DLayer = derive('TransposedConv1DLayer').based_on(TransposedConvLayer).with_fixed(ndim=1)
TransposedConv2DLayer = derive('TransposedConv2DLayer').based_on(TransposedConvLayer).with_fixed(ndim=2)
TransposedConv3DLayer = derive('TransposedConv3DLayer').based_on(TransposedConvLayer).with_fixed(ndim=3)

transposed_conv_1d = model_from(TransposedConv1DLayer)()
deconv_1d = transposed_conv_1d

transposed_conv_2d = model_from(TransposedConv2DLayer)()
deconv_2d = transposed_conv_2d

transposed_conv_3d = model_from(TransposedConv3DLayer)()
deconv_3d = transposed_conv_3d

transposed_conv = select_by_dim([None, transposed_conv_1d, transposed_conv_2d, transposed_conv_3d])
deconv = transposed_conv

PaddedTransposedConv1DLayer = derive('PaddedTransposedConv1DLayer').based_on(TransposedConvLayer).with_fixed(ndim=1, pad='same')
PaddedTransposedConv2DLayer = derive('PaddedTransposedConv2DLayer').based_on(TransposedConvLayer).with_fixed(ndim=2, pad='same')
PaddedTransposedConv3DLayer = derive('PaddedTransposedConv3DLayer').based_on(TransposedConvLayer).with_fixed(ndim=3, pad='same')

ptransposed_conv_1d = model_from(PaddedTransposedConv1DLayer)()
pdeconv_1d = ptransposed_conv_1d

ptransposed_conv_2d = model_from(PaddedTransposedConv2DLayer)()
pdeconv_2d = ptransposed_conv_2d

ptransposed_conv_3d = model_from(PaddedTransposedConv3DLayer)()
pdeconv_3d = ptransposed_conv_3d

ptransposed_conv = select_by_dim([None, ptransposed_conv_1d, ptransposed_conv_2d, ptransposed_conv_3d])
pdeconv = ptransposed_conv


import matplotlib.pyplot as plt
import numpy as np
import os.path as osp

__all__ = [
  'SimpleWatcher',
  'ExperimentWatcher'
]

try:
  from IPython import display
except ImportError:
  display = None

### http://colorbrewer2.org/#type=qualitative&scheme=Set1&n=8
_default_color_seq = ('#377eb8', '#e41a1c', '#4daf4a', '#984ea3', '#ff7f00', '#a65628', '#f781bf', '#999999')

class SimpleWatcher(object):
  def __init__(self, title='Watcher', labels=('loss', ), colors=_default_color_seq, figsize=(9, 6), save_dir='./'):
    self.save_dir = save_dir

    self.fig = plt.figure(figsize=figsize)
    self.ax = self.fig.add_subplot(111)

    self.ax.set_xlim([0.0, 1.0])
    self.ax.set_ylim([0.0, 1.0])

    self.fig.suptitle(title)
    self.title = title

    self.colors = colors
    self.labels = labels
    self.drawn = False

    self.data = None
    self.reset()

  def reset(self):
    self.data = [
      ([], [], [], [], []) for _ in range(len(self.labels))
    ]

  def _get_ylim(self):
    min_y = min([min(x_min) for x_min, _, _, _, _ in self.data])
    max_y = max([max(x_max) for _, _, _, _, x_max in self.data])

    delta = (max_y - min_y) / 2
    center = (max_y + min_y) / 2

    lower_bound = center - 1.05 * delta
    upper_bound = center + 1.05 * delta

    return lower_bound, upper_bound

  def update(self, *data):
    for d, new_data in zip(self.data, data):
      lower2, lower1, upper1, upper2 = np.percentile(new_data, q=(5, 20, 80, 95), axis=0)
      mean = np.mean(new_data)

      d[0].append(lower2)
      d[1].append(lower1)
      d[2].append(mean)
      d[3].append(upper1)
      d[4].append(upper2)

    return self

  def draw(self):
    self.ax.clear()
    display.clear_output(wait=True)

    x_lim = max([len(d[0]) for d in self.data])
    self.ax.set_xlim(0.0, x_lim)

    y_lower, y_upper = self._get_ylim()
    self.ax.set_ylim([y_lower, y_upper])

    for d, color, label in zip(self.data, self.colors, self.labels):
      lower2, lower1, mean, upper1, upper2 = d

      iters = np.arange(len(d[0])) + 0.5

      self.ax.fill_between(iters, lower1, upper1, alpha=0.2, color=color)
      self.ax.fill_between(iters, lower2, upper2, alpha=0.1, color=color)

      self.ax.plot(iters, mean, label=label, color=color)

    if not self.drawn:
      self.ax.legend()
      self.drawn = False

    display.display(self.fig)

    self.fig.savefig(osp.join(self.save_dir, '%s.png' % self.title), dpi=120)

class ExperimentWatcher(object):
  def __init__(self, title='learning curves', colors=_default_color_seq, figsize=(9, 6), save_path=None):
    self.save_path = save_path

    self.fig = plt.figure(figsize=figsize)
    self.ax = self.fig.add_subplot(111)

    self.ax.set_xlim([0.0, 1.0])
    self.ax.set_ylim([0.0, 1.0])

    self.fig.suptitle(title)
    self.title = title
    self.colors = colors

    self.drawn = False

    self.data = None

  def reset(self):
    self.data = None

  def _get_ylim(self):
    min_y = min([
      min(d) for name in self.data for d in self.data[name]
    ])

    max_y = max([
      max(d) for name in self.data for d in self.data[name]
    ])

    delta = (max_y - min_y) / 2
    center = (max_y + min_y) / 2

    lower_bound = center - 1.05 * delta
    upper_bound = center + 1.05 * delta

    return lower_bound, upper_bound

  def update(self, epoch_history):
    if self.data is None:
      self.data = dict([
        (name, [
          [] for _ in epoch_history[name]
        ]) for name in epoch_history
      ])

    for name in epoch_history:
      for history, d in zip(epoch_history[name], self.data[name]):
        d.append(np.mean(history))

    return self

  def draw(self):
    if self.data is None:
      return

    self.ax.clear()
    display.clear_output(wait=True)

    x_lim = max([
      max([ len(d) for d in self.data[name] ])
      for name in self.data
    ])
    self.ax.set_xlim(0.0, x_lim)

    y_lower, y_upper = self._get_ylim()
    self.ax.set_ylim([y_lower, y_upper])

    for name, color in zip(self.data, self.colors):
      for i, mean in enumerate(self.data[name]):
        iters = np.arange(len(mean)) + 0.5

        if i == 0:
          self.ax.plot(iters, mean, label=name, color=color)
        else:
          self.ax.plot(iters, mean, color=color)

    if not self.drawn:
      self.ax.legend()
      self.drawn = False

    display.display(self.fig)

    if self.save_path is not None:
      self.fig.savefig(self.save_path, dpi=120)


import numpy as np
import tensorflow as tf

from .meta import Dataset
from .variable import VariableDataset

__all__ = [
  'SlicedSubset',
  'IndexedSubset'
]

class SlicedSubset(VariableDataset):
  def __init__(self, dataset : Dataset, item : slice or int):
    super(SlicedSubset, self).__init__(*[
      d[item] for d in dataset.data()
    ])

class IndexedSubset(Dataset):
  def __init__(self, dataset : Dataset, indx):
    self.indx = tf.convert_to_tensor(indx)
    self.dataset = dataset

    super(IndexedSubset, self).__init__()

  def _get_single(self, item):
    return self.dataset._get_single(self.indx[item])

  def _get_slice(self, item):
    indx = self.indx[item]

    return self.dataset._get_sparse(indx)

  def _get_sparse(self, item):
    indx = tf.gather(self.indx, indices=item)
    return self.dataset._get_sparse(indx)

  def get_subset(self, item):
    if isinstance(item, slice):
      return IndexedSubset(self.dataset, self.indx[item])
    else:
      indx = tf.gather(self.indx, item)
      return IndexedSubset(self.dataset, indx)

  def size(self):
    return tf.shape(self.indx)[0]

  def data(self):
    return self.dataset._get_sparse(self.indx)

  def shapes(self):
    size = self.size()
    return tuple(
      (size, ) + shape[1:]
      for shape in self.dataset.shapes()
    )


import tensorflow as tf
import collections

from .meta import Dataset

class VariableDataset(Dataset):
  def __init__(self, *variables):
    self.variables = tuple(variables)

    super(VariableDataset, self).__init__()

  @classmethod
  def _slice_class(cls):
    from .subsets import SlicedSubset
    return SlicedSubset

  def _get_single(self, item):
    return tuple(
      var[item]
      for var in self.variables
    )

  def _get_slice(self, item):
    return tuple(
      var[item]
      for var in self.variables
    )

  def _get_sparse(self, item):
    return tuple(
      tf.gather(var, item)
      for var in self.variables
    )

  def size(self):
    return tf.shape(self.variables[0])[0]

  @tf.function
  def assign(self, *data):
    return [
      var.assign(value)
      for var, value in zip(self.variables, data)
    ]

  def data(self):
    return self.variables

  def shapes(self):
    return tuple(
      var.shape for var in self.variables
    )

def empty_dataset(*shapes, dtypes=tf.float32):
  assert len(shapes) > 0

  if isinstance(dtypes, str) or isinstance(dtypes, tf.dtypes.DType):
    dtypes = [dtypes for _ in shapes]

  shapes = [
    shape_or_array.shape if hasattr(shape_or_array, 'shape') else shape_or_array
    for shape_or_array in shapes
  ]

  general_shapes = [
    (None,) + shape[1:]
    for shape in shapes
  ]

  if not isinstance(dtypes, collections.Iterable) or isinstance(dtypes, str):
    default_dtype = dtypes
    dtypes = []

    for shape_or_array in shapes:
      if hasattr(shape_or_array, 'dtype'):
        dtypes.append(shape_or_array.dtype)
      else:
        dtypes.append(default_dtype)

  variables = [
    tf.Variable(
      initial_value=tf.zeros(shape, dtype=dtype),
      dtype=dtype,
      validate_shape=False,
      shape=general_shape,
    ) for shape, general_shape, dtype in zip(shapes, general_shapes, dtypes)
  ]

  return VariableDataset(*variables)


def variable_dataset(*data):
  dataset = empty_dataset(
    *(d.shape for d in data),
    dtypes=tuple(str(d.dtype) for d in data)
  )
  dataset.assign(*data)
  return dataset